# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""This file defines various classes and helper functions for working
with results."""

from collections import OrderedDict
import re

import wikitcms.helpers

# These are used by multiple functions, so let's share them.
# Wiki table row separator
SEP_PATT = re.compile(r"\|[-\}].*?\n")
# Identifies an instance of the result template. Will break if the
# result contains another template, but don't do that. the lookahead
# is used to capture the 'comments' for the result: we keep matching
# until we hit the next instance of the template, or a cell or row
# separator (newline starting with a |). The re.S is vital.
RES_PATT = re.compile(r"{{result.+?}}.*?(?=\n*{{result|$|\n\|)", re.S)


def _filter_results(results, statuses=None, transferred=True, bot=True):
    """Filter results. Shared between next two functions."""
    # Drop example / sample results
    results = [
        r
        for r in results
        if not r.user
        or r.user.lower() not in ("sampleuser", "exampleuser", "example", "username", "fasname")
    ]
    if statuses:
        results = [r for r in results for s in statuses if r.status and s in r.status]
    if not transferred:
        results = [r for r in results if not r.transferred]
    if not bot:
        results = [r for r in results if not r.bot]
    return results


def find_results(text, statuses=None, transferred=True, bot=True):
    """Find test results in a given chunk of wiki text. Returns a list
    of Result objects. If statuses is not None, it should be an
    iterable of strings, and only result objects whose status matches
    one of the given statuses will be returned. If transferred is
    False, results like {{result|something|previous TC5 run}} will not
    be included. If bot is False, results from automated test bots
    (results with 'bot=true') will not be included.
    """
    results = list()
    # Identifies an instance of the old {{testresult template, in the
    # same way as RES_PATT (above).
    oldres_patt = re.compile(r"{{testresult.+?}}.*?(?={{testresult|$)", re.M)
    for res in RES_PATT.findall(text):
        results.append(Result.from_result_template(res))
    for oldres in oldres_patt.findall(text):
        results.append(Result.from_testresult_template(oldres))

    results = _filter_results(results, statuses, transferred, bot)
    return results


def find_results_by_row(text, statuses=None):
    """Find test results using a row-by-row scan, guessing the user
    for results which do not have one. Used for Test Day pages. Note:
    doesn't bother handling {{testresult because AFAICT no Test Day
    pages use that template.
    """
    # Slightly slapdash way to identify the contents of a wiki table
    # cell. Good enough to find the contents of the first cell in the
    # row, which we do because it'll usually be the user name in a
    # Test Day results table.
    cell_patt = re.compile(r"\| *(.*?) *\n\|")
    # Captures the user name from a wikilink to a user page (typically
    # used to indicate the user who reports a result).
    user_patt = re.compile(r"\[\[User: *(.*?) *[|\]]")
    results = list()

    for row in SEP_PATT.split(text):
        for match in RES_PATT.findall(row):
            res = Result.from_result_template(match)
            if res.status and not res.user:
                # We'll try and find a [[User: wikilink; if we can't,
                # we'll guess that the reporter's name is the contents
                # of the first cell in the row.
                pattuser = user_patt.search(row)
                if pattuser:
                    res.user = pattuser.group(1).lower()
                else:
                    celluser = cell_patt.search(row)
                    if celluser:
                        res.user = celluser.group(1).lower()
            results.append(res)

    results = _filter_results(results, statuses)
    return results


def find_resultrows(text, section="", secid=0, statuses=None, transferred=True):
    """Find result rows in a given chunk of wiki text. Returns a list
    of ResultRow objects. 'statuses' and 'transferred' are passed all
    the way through ResultRow to find_results() and behave as
    described there, for the Result objects in each ResultRow.
    """
    # oh shut up pylint what do you have against variables
    # pylint: disable=too-many-locals

    # identify all test case names, including old ones. modern ones
    # match QA:Testcase.*, but older ones sometimes have QA/TestCase.
    testcase_pattern = re.compile(r"(QA[:/]Test.+?)[\|\]\n]")
    # row separator is |-, end of table is |}
    columns = list()
    resultrows = list()
    rows = SEP_PATT.split(text)
    for row in rows:
        rowlines = row.split("\n")
        for line in rowlines:
            # check if this is a column header row, and update column
            # names. Sometimes the header row doesn't have an explicit
            # row separator so the 'row' might be polluted with
            # preceding lines, so we split the row into lines and
            # check each line in the row.
            line = line.strip()
            if line.find("!") == 0 and line.find("!!") > 0:
                # column titles. note: mw syntax in fact allows for
                # '! title\n! title\n! title' as well as '! title !!
                # title !! title'. But we don't use that syntax.
                columns = line.lstrip("!").split("!!")
                for column in columns:
                    # sanitize names a bit
                    newcol = column.strip()
                    newcol = newcol.strip("'[]")
                    newcol = newcol.strip()
                    try:
                        # drop out any <ref> block
                        posa = newcol.index("<ref>")
                        posb = newcol.index("</ref>") + 6  # length
                        newcol = newcol[:posa] + newcol[posb:]
                        newcol = newcol.strip()
                    except ValueError:
                        pass
                    try:
                        newcol = newcol.split("|")[1]
                    except IndexError:
                        pass
                    if newcol != column:
                        columns.insert(columns.index(column), newcol)
                        columns.remove(column)
        tcmatch = testcase_pattern.search(row)
        if tcmatch:
            # *may* be a result row - may also be a garbage 'row'
            # between tables which happens to contain a test case
            # name. So we get a ResultRow object but discard it if it
            # doesn't contain any result cells. This test works even
            # if the actual results are filtered by statuses= or
            # Transferred=, because the resrow.results dict will
            # always have a key for each result column, though its
            # value may be an empty list.
            resrow = ResultRow.from_wiki_row(
                tcmatch.group(1), columns, row, section, secid, statuses, transferred
            )
            if resrow.results:
                resultrows.append(resrow)
    return resultrows


class Result:
    """A class that represents a single test result. Note that a
    'none' result, as you get if you just instantiate this class
    without arguments, is a thing, at least for wikitcms; when text
    with {{result|none}} templates in it is parsed, such objects may
    be created/returned, and you can produce the {{result|none}} text
    as the result_template property of such an instance.

    You would usually instantiate this class directly to report a new
    result.

    Methods that parse existing results will use one of the class
    methods that returns a Result() with the appropriate attributes.
    When one of those parsers produces an instance it will pass
    origtext to record the exact string parsed to produce the instance;
    otherwise origtext should be left blank.

    transferred, if True, indicates the result is of the "previous
    (compose) run" type that is used to indicate where we think a
    result from a previous compose is valid for a later one.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, status=None, user=None, bugs=None, comment="", bot=False, origtext=""):
        # pylint: disable=too-many-arguments
        self.status = status
        self.user = user
        self.bugs = bugs
        if self.bugs:
            self.bugs = [str(bug) for bug in self.bugs]
        self.comment = comment
        self.bot = bot
        self.transferred = False
        self.comment_bugs = wikitcms.helpers.find_bugs(self.comment)
        self.origtext = origtext

    def __str__(self):
        if not self.status:
            return "Result placeholder - {{result|none}}"
        if self.bot:
            bot = "BOT "
        else:
            bot = ""
        status = "Result: " + self.status.capitalize()
        if self.transferred:
            user = " transferred: " + self.user
        elif self.user:
            user = " from " + self.user
        else:
            user = ""
        if self.bugs:
            bugs = ", bugs: " + ", ".join(self.bugs)
        else:
            bugs = ""
        if self.comment:
            comment = ", comment: " + self.comment
            # Don't display ref tags
            refpatt = re.compile(r"</?ref.*?>")
            comment = refpatt.sub("", comment)
        else:
            comment = ""
        return bot + status + user + bugs + comment

    @property
    def result_template(self):
        """The {{result}} template string that would represent the
        properties of this result in a wiki page.
        """
        bugtext = ""
        commtext = self.comment
        usertext = ""
        bottext = ""
        if self.status is None:
            status = "none"
        else:
            status = self.status
        if self.bugs:
            bugtext = "|" + "|".join(self.bugs)
        if self.user:
            usertext = "|" + self.user
        if self.bot:
            bottext = "|bot=true"
        return f"{{{{result|{status}{usertext}{bugtext}{bottext}}}}}{commtext}"

    @classmethod
    def from_result_template(cls, string):
        """Returns a Result object based on the {{result}} template.
        The most complex result template you see might be:
        {{ result | fail| bot =true| adamwill | 123456|654321|615243}} comment
        We want the 'fail' and 'adamwill' bits separately and stripped,
        and all the bug numbers in one chunk to be parsed later to
        construct a list of bugs, and none of the pipes, brackets, or
        whitespace. Mediawiki named parameters can occur anywhere in
        the template and aren't counted in the numbered parameters, so
        we need to find them and extract them first. We record the
        comment exactly as is.
        """
        # pylint: disable=too-many-locals
        template, comment = string.strip().split("}}", 1)
        comment = comment.strip()
        template = template.lstrip("{")
        params = template.split("|")
        namedpars = dict()
        bot = False

        for param in params:
            if "=" in param:
                (par, val) = param.split("=", 1)
                namedpars[par.strip()] = val.strip()
                params.remove(param)
        if "bot" in namedpars and namedpars["bot"]:
            # This maybe doesn't do what you expect for 'bot=false',
            # but we don't handle that in Mediawiki either and we want
            # to stay consistent.
            bot = True

        # 'params' now contains only numbered params
        # Pad the non-existent parameters to make things cleaner later
        while len(params) < 3:
            params.append("")

        for i, param in enumerate(params):
            params[i] = param.strip()
            if params[i] == "":
                params[i] = None
        status, user = params[1:3]
        bugs = params[3:]
        if status and status.lower() == "none":
            status = None

        if bugs:
            bugs = [b.strip() for b in bugs if b and b.strip()]
            for i, bug in enumerate(bugs):
                # sometimes people write 123456#c7, remove the suffix
                if "#" in bug:
                    newbug = bug.split("#")[0]
                    if newbug.isdigit():
                        bugs[i] = newbug

        res = cls(status, user, bugs, comment, bot, origtext=string)
        if user and "previous " in user:
            res.transferred = True
        return res

    @classmethod
    def from_testresult_template(cls, string):
        """Returns a Result object based on the {{testresult}} template.
        This was used in Fedora 12. It looks like this:
        {{testresult/pass|FASName}} <ref>comment or bug</ref>
        The bug handling here is very special-case - it relies on the
        fact that bug IDs were always six-digit strings, at the time,
        and on the template folks used to link to bug reports - but
        should be good enough.
        """
        bug_patt = re.compile(r"({{bz.*?(\d{6,6}).*?}})")
        emptyref_patt = re.compile(r"<ref> *?</ref>")
        template, comment = string.strip().split("}}", 1)
        template = template.lstrip("{")
        template = template.split("/")[1]
        params = template.split("|")
        try:
            status = params[0].strip().lower()
            if status == "none":
                status = None
        except IndexError:
            status = None
        try:
            user = params[1].strip().lower()
        except IndexError:
            user = None
        bugs = [b[1] for b in bug_patt.findall(comment)]
        if comment:
            comment = bug_patt.sub("", comment)
            comment = emptyref_patt.sub("", comment)
            if comment.replace(" ", "") == "":
                comment = ""
            comment = comment.strip()
        else:
            pass
        res = cls(status, user, bugs, comment, origtext=string)
        if user and "previous " in user:
            res.transferred = True
        return res


class ResultRow:
    """Represents the 'test instance' concept for Wikitcms, where it
    is a result row from the tables used to contain results. We think
    of a "test instance" as the 'basic unit' of a single test for
    which some results are expected to be reported.

    The 'testcase' is the basic identifier of a ResultRow. It will
    not necessarily be unique, though - it is not uncommon for a set
    of ResultRows to have the same 'testcase' but a different 'name',
    which is the link text: there will a column which for each row
    contains [[testcase|name]], the testcase being the same but the
    name being different. Commonly the 'testcase' and 'name' will be
    the same, when each instance within a set has a different
    'testcase' the name should be identical to the testcase.

    milestone is, roughly, the priority of the test: it will usually
    be Basic, Beta or Final.

    results is required to be a dict of lists; the dict keys represent
    the test's environments. The values must be lists of instances of
    wikitcms.Result or a subclass of it.

    columns is the list of columns in the table in which the result
    was found (this is needed to figure out the environments for the
    results, as the envs are represented by table columns, and to know
    which cell to edit when modifying results).

    origtext is the text which was parsed to produce the instance, if
    it was produced by the from_wiki_row() class method which parses
    wiki text to produce instances.

    section and secid are the wiki page section in which the table
    from which the row came is located; though these are in a way
    attributes of the page, this is really another case where an MW
    attribute is just a way of encoding information about a *test*.
    The splitting of result pages into sections is a way of sub-
    grouping tests in each page. So it's appropriate to store those
    attributes here.

    At present you typically get ResultRow instances by calling a
    ComposePage's get_resultrows() method, which runs its text through
    result.find_resultrows(), which isolates the result rows in the
    wiki text and runs through through this class' from_wiki_row()
    method. This will always provide instances with a full set of
    the above-described attributes.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self, testcase, columns, section="", secid=None, milestone="", origtext="", results=None
    ):
        # pylint: disable=too-many-arguments
        self.testcase = testcase
        self.name = testcase
        self.milestone = milestone
        if not results:
            self.results = dict()
        else:
            self.results = results
        self.columns = columns
        self.origtext = origtext
        self.section = section
        self.secid = secid

    def matches(self, other):
        """This is roughly an 'equals' check for ResultRows; if all
        identifying characteristics and the origtext match, this is
        True, otherwise False. __dict__ match is too strong as the
        'results' attribute is a list of Result instances; each time
        you instantiate the same ResultRow you get different Result
        objects. We don't override __eq__ and use == because that has
        icky implications for hashing, and we just want to stay away
        from that mess:
        https://docs.python.org/3/reference/datamodel.html#object.__hash__
        """
        if isinstance(other, self.__class__):
            ours = (self.testcase, self.name, self.secid, self.origtext)
            theirs = (other.testcase, other.name, other.secid, other.origtext)
            return ours == theirs
        return False

    @classmethod
    def from_wiki_row(
        cls, testcase, columns, text, section, secid, statuses=None, transferred=True
    ):
        """Instantiate a ResultRow from some wikitext and some info
        that is worked out from elsewhere in the page.
        """
        # pylint: disable=too-many-locals, too-many-arguments
        results = OrderedDict()
        # this is presumptuous, but holds up for every result page
        # tested so far; there may be some with whitespace, and
        # '| cell || cell || cell' is permitted as an alternative to
        # '| cell\n| cell\n| cell' but we do not seem to use it.
        cells = text.split("\n|")
        milestone = ""
        for mile in ("Alpha", "Basic", "Beta", "Final", "Optional", "Tier1", "Tier2", "Tier3"):
            if mile in cells[0]:
                milestone = mile
                # we take the *first* milestone we find, so we treat
                # e.g. "Basic / Final" as Basic
                break
        for i, cell in enumerate(cells):
            if testcase in cell:
                try:
                    # see if we can find some link text for the test
                    # case, and assume it's the test's "name" if so
                    altname = cell.strip().strip("[]").split("|")[1]
                    continue
                except IndexError:
                    try:
                        altname = cell.strip().strip("[]").split(maxsplit=1)[1]
                    except IndexError:
                        altname = None
            if "{{result" in cell or "{{testresult" in cell:
                # any cell containing a result string is a 'result
                # cell', and the index of the cell in columns will be
                # the title of the column it is in. find_results()
                # returns an empty list if all results are filtered
                # out, so the results dict's keys will always
                # represent the full set of environments for this test
                try:
                    results[columns[i]] = find_results(cell, statuses, transferred)
                except IndexError:
                    # FIXME: log (messy table, see e.g. F15 'Multi
                    # Image')
                    pass
        row = cls(testcase, columns, section, secid, milestone, text, results)
        if altname:
            row.name = altname
        return row


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Classes that describe distribution releases are defined here."""

import wikitcms.page
import wikitcms.listing


class Release:
    """Class for a Fedora release. wiki is a wikitcms site object.
    Release is a string containing a Fedora release version (e.g. 21).
    """

    def __init__(self, release, wiki, dist="Fedora"):
        self.release = release
        self.dist = dist
        self.category_name = f"Category:{self.dist} {self.release} Test Results"
        self.site = wiki

    @property
    def testday_pages(self):
        """All Test Day pages for this release (as a list)."""
        cat = self.site.pages[f"Category:Fedora {self.release} Test Days"]
        return [page for page in cat if isinstance(page, wikitcms.page.TestDayPage)]

    def milestone_pages(self, milestone=None):
        """If no milestone, will give all release validation pages for
        this release (as a generator). If a milestone is given, will
        give validation pages only for that milestone. Note that this
        works by category; you may get somewhat different results by
        using page name prefixes.
        """
        cat = wikitcms.listing.ValidationCategory(
            self.site, self.release, milestone, dist=self.dist
        )
        pgs = self.site.walk_category(cat)
        return (p for p in pgs if isinstance(p, wikitcms.page.ValidationPage))


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2016 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument, too-many-arguments
# pylint: disable=too-few-public-methods

"""Tests for wiki.py."""

from unittest import mock

import openidc_client
import openidc_client.requestsauth
import pytest

import wikitcms.wiki as wk
import wikitcms.event
import wikitcms.result

FAKE_CURRENT_COMPOSE = """{{tempdoc}}
<onlyinclude>{{#switch: {{{1|full}}}
| full = 24 Alpha 1.1
| release = 24
| milestone = Alpha
| compose = 1.1
| date =
}}</onlyinclude>
[[Category: Fedora Templates]]
"""


class FakeRow:
    """This is a very bare fake ResultRow class; we need it to fully
    test report_validation_results, can't really do it with Mocks. We
    use mock to replace the 'real' page.find_resultrow method with the
    fake_findrow method below, which returns instances of this class.
    """

    def __init__(self, name):
        self.name = name

    def matches(self, other):
        """Simple matches replacement."""
        return self.name == other.name


def fake_findrow(self, testcase="", section="", testname="", env=""):
    """This is a fake find_resultrow function which just returns
    FakeRows based on the values it's given. We use mock to replace
    Page instances' find_resultrow methods with this, further down.
    Note we're simulating a case where the found row covers both
    reported envs here; find_resultrow can also cover a case where
    there are two rows with equal case, section and name but each
    covering different envs, but we'll test that in its own test.
    """
    name = ".".join((testcase, section, testname))
    return FakeRow(name)


def fake_pginit1(self, site, release, testtype, milestone, compose, info=None, dist="Fedora"):
    """Fake page.__init__ for testing get_validation_page guessing.
    This one gives us pages that 'exist' if they're Rawhide and don't
    otherwise.
    """
    self.milestone = milestone
    if milestone == "Rawhide":
        self.exists = True
    else:
        self.exists = False


def fake_pginit2(self, site, release, testtype, milestone, compose, info=None, dist="Fedora"):
    """Fake page.__init__ for testing get_validation_page guessing.
    This one gives us pages that 'exist' if they're Branched and don't
    otherwise.
    """
    self.milestone = milestone
    if milestone == "Branched":
        self.exists = True
    else:
        self.exists = False


def fake_pginit3(self, site, release, testtype, milestone, compose, info=None, dist="Fedora"):
    """Fake page.__init__ for testing get_validation_page guessing.
    This one just never exists.
    """
    self.exists = False


class TestWiki:
    """Tests for the functions in wiki.py."""

    # the 'fjajah' is just to make sure we're running offline; if I
    # screw up and add a test that hits the network it'll cause the
    # tests to hang/fail instead of succeeding a bit slower
    site = wk.Wiki("fjajah", do_init=False, force_login=False)

    @mock.patch("mwclient.page.Page.__init__", return_value=None)
    @mock.patch("mwclient.page.Page.text", return_value=FAKE_CURRENT_COMPOSE)
    def test_current_compose(self, faketext, fakeinit):
        """Tests for current_compose and current_modular_compose."""
        assert self.site.current_compose == {
            "full": "24 Alpha 1.1",
            "release": "24",
            "milestone": "Alpha",
            "compose": "1.1",
            "date": "",
        }
        assert self.site.current_modular_compose == {
            "full": "24 Alpha 1.1",
            "release": "24",
            "milestone": "Alpha",
            "compose": "1.1",
            "date": "",
        }

    @mock.patch("mwclient.page.Page.__init__", return_value=None)
    @mock.patch("mwclient.page.Page.text", return_value=FAKE_CURRENT_COMPOSE)
    @mock.patch("wikitcms.wiki.Wiki.get_validation_event", autospec=True)
    def test_current_event(self, fakeget, faketext, fakeinit):
        """Tests for current_event and current_modular_event."""
        _ = self.site.current_event
        fakeget.assert_called_with(
            mock.ANY, compose="1.1", milestone="Alpha", release="24", dist="Fedora"
        )
        fakeget.reset_mock()
        _ = self.site.current_modular_event
        fakeget.assert_called_with(
            mock.ANY, compose="1.1", milestone="Alpha", release="24", dist="Fedora-Modular"
        )

    @mock.patch("mwclient.page.Page.__init__", return_value=None)
    @mock.patch("mwclient.page.Page.text", return_value="foobar")
    @mock.patch("mwclient.page.Page.save")
    def test_add_to_category(self, fakesave, faketext, fakeinit):
        """Test for add_to_category."""
        self.site.add_to_category("Foobar", "Category:Some category", "summary")
        fakesave.assert_called_with(
            "foobar\n[[Category:Some category]]", "summary", createonly=False
        )
        # now check case where page is already in the category
        faketext.return_value = "foobar\n[[Category:Some category]]"
        fakesave.reset_mock()
        self.site.add_to_category("Foobar", "Category:Some category", "summary")
        # we should return without calling save
        assert fakesave.call_count == 0

    def test_walk_category(self, fakemwp):
        """Test for walk_category."""
        # we have to mock up a whole nested category structure here
        # we're gonna walk 'fakecat1', which contains two regular
        # pages and 'fakecat2', which contains another regular page
        fakecat1 = self.site.pages["Category:Somecat"]
        fakecat1.namespace = 14
        fakecat2 = self.site.pages["Category:Someothercat"]
        fakecat1.namespace = fakecat2.namespace = 14
        fakepg1 = self.site.pages["Somepage1"]
        fakepg2 = self.site.pages["Somepage2"]
        fakepg3 = self.site.pages["Somepage3"]
        fakepg1.namespace = fakepg2.namespace = fakepg3.namespace = 1
        # this isn't really a perfect test, but it's the best I can
        # come up with for now
        with mock.patch("mwclient.listing.GeneratorList.__next__") as fakenext:
            fakenext.side_effect = [fakepg1, fakepg2, fakecat2, fakepg3]
            assert list(self.site.walk_category(fakecat1)) == [fakepg1, fakepg2, fakepg3]

    @mock.patch("fedfind.release.Compose.exists", return_value=True)
    @mock.patch("fedfind.release.IoTNightly.label", "RC-20200513.0")
    @mock.patch("fedfind.release.IoTNightly.cid", "Fedora-IoT-33-20200513.0")
    @mock.patch("fedfind.release.Production.label", "RC-1.6")
    @mock.patch("fedfind.release.Production.cid", "Fedora-27-20171105.0")
    @mock.patch("fedfind.helpers.get_current_release", autospec=True, return_value=27)
    @mock.patch("mwclient.page.Page.__init__", return_value=None)
    @mock.patch("mwclient.page.Page.text", return_value=FAKE_CURRENT_COMPOSE)
    @mock.patch("wikitcms.event.NightlyEvent", autospec=True)
    @mock.patch("wikitcms.event.ComposeEvent", autospec=True)
    def test_get_validation_event(
        self, fakecompose, fakenightly, faketext, fakeinit, fakegetcurr, fakecompexists
    ):
        """Various tests for get_validation_event."""
        # current event
        self.site.get_validation_event()
        fakecompose.assert_called_with(self.site, "24", "Alpha", "1.1", dist="Fedora", cid="")
        self.site.get_validation_event(dist="Fedora-Modular")
        fakecompose.assert_called_with(
            self.site, "24", "Alpha", "1.1", dist="Fedora-Modular", cid=""
        )
        # old-school TC/RC
        self.site.get_validation_event(23, "Final", "TC9")
        fakecompose.assert_called_with(self.site, 23, "Final", "TC9", dist="Fedora", cid="")
        self.site.get_validation_event(23, "Beta", "RC1")
        fakecompose.assert_called_with(self.site, 23, "Beta", "RC1", dist="Fedora", cid="")
        # old-school nightly
        self.site.get_validation_event(23, "Rawhide", "20151112")
        fakenightly.assert_called_with(
            self.site, release=23, milestone="Rawhide", compose="20151112", dist="Fedora"
        )
        self.site.get_validation_event(23, "Branched", "20151211", dist="Fedora")
        fakenightly.assert_called_with(
            self.site, release=23, milestone="Branched", compose="20151211", dist="Fedora"
        )
        # Pungi 4 production/candidate
        self.site.get_validation_event(24, "Alpha", "1.1")
        fakecompose.assert_called_with(self.site, 24, "Alpha", "1.1", dist="Fedora", cid="")
        self.site.get_validation_event(27, "Beta", "1.5", dist="Fedora-Modular")
        fakecompose.assert_called_with(self.site, 27, "Beta", "1.5", dist="Fedora-Modular", cid="")
        # Past 23, 'Final' milestone should be converted to 'RC'
        self.site.get_validation_event(25, "Final", "1.1")
        fakecompose.assert_called_with(self.site, 25, "RC", "1.1", dist="Fedora", cid="")
        # Pungi 4 nightly
        self.site.get_validation_event(24, "Rawhide", "20160222.n.0")
        fakenightly.assert_called_with(
            self.site, release=24, milestone="Rawhide", compose="20160222.n.0", dist="Fedora"
        )
        self.site.get_validation_event(24, "Branched", "20160315.n.1")
        fakenightly.assert_called_with(
            self.site, release=24, milestone="Branched", compose="20160315.n.1", dist="Fedora"
        )
        self.site.get_validation_event(27, "Branched", "20171110.n.1", dist="Fedora-Modular")
        fakenightly.assert_called_with(
            self.site,
            release=27,
            milestone="Branched",
            compose="20171110.n.1",
            dist="Fedora-Modular",
        )
        # Rawhide nightly compose ID
        self.site.get_validation_event(cid="Fedora-Rawhide-20180220.n.0", dist="Fedora")
        fakenightly.assert_called_with(
            self.site, release="28", milestone="Rawhide", compose="20180220.n.0", dist="Fedora"
        )
        # Branched nightly compose ID
        self.site.get_validation_event(cid="Fedora-27-20171120.n.0", dist="Fedora")
        fakenightly.assert_called_with(
            self.site, release="27", milestone="Branched", compose="20171120.n.0", dist="Fedora"
        )
        # Candidate compose ID (note compose ID passthrough)
        self.site.get_validation_event(cid="Fedora-27-20171105.0", dist="Fedora")
        fakecompose.assert_called_with(
            self.site, "27", "RC", "1.6", dist="Fedora", cid="Fedora-27-20171105.0"
        )
        # IoT compose: these are technically productions but quack and
        # walk like nightlies, with dates in the labels; we intend to
        # treat them as nightlies and create nightly events
        fakegetcurr.return_value = 32
        self.site.get_validation_event(cid="Fedora-IoT-33-20200513.0", dist="Fedora-IoT")
        fakenightly.assert_called_with(self.site, "33", "RC", "20200513.0", dist="Fedora-IoT")

        with pytest.raises(ValueError):
            # Non-nightly compose but no milestone
            self.site.get_validation_event(24, "", "1.1")
        # Invalid composes
        with pytest.raises(ValueError):
            self.site.get_validation_event(24, "Branched", "foobar")
        with pytest.raises(ValueError):
            self.site.get_validation_event(24, "Branched", "TC1a")
        with pytest.raises(ValueError):
            self.site.get_validation_event(24, "Branched", "1.1a")
        # looks kinda like a date but is not one
        with pytest.raises(ValueError):
            self.site.get_validation_event(24, "Branched", "20161356")
        with pytest.raises(ValueError):
            self.site.get_validation_event(24, "Branched", "20161356.n.0")
        with pytest.raises(ValueError):
            # invalid type
            self.site.get_validation_event(24, "Branched", "20160314.x.0")

    @mock.patch("wikitcms.wiki.Wiki.allresults")
    @mock.patch("mwclient.page.Page.text", return_value="foobar")
    def test_get_validation_event_guess_nightly(self, faketext, fakeresults, fakemwp):
        """Test get_validation_event guessing milestone for nightly
        events.
        """
        fakepage = self.site.pages["Test Results:Fedora 24 Rawhide 20160314.n.0 Base"]
        otherpage = self.site.pages["Somepage"]
        # this will test the Rawhide path, as the *first* fakepages
        # call will hit
        fakeresults.side_effect = ([fakepage],)
        ret = self.site.get_validation_event(24, "", "20160314.n.0")
        assert isinstance(ret, wikitcms.event.NightlyEvent)
        assert ret.milestone == "Rawhide"
        # this will test the Branched path, as the *second* fakepages
        # call will hit
        fakeresults.side_effect = ([otherpage], [fakepage])
        ret = self.site.get_validation_event(24, "", "20160314.n.0")
        assert isinstance(ret, wikitcms.event.NightlyEvent)
        assert ret.milestone == "Branched"
        # ...and this won't hit anywhere and should raise
        fakeresults.side_effect = ([otherpage], [otherpage])
        with pytest.raises(ValueError) as err:
            self.site.get_validation_event(24, "", "20160314.n.0")
            assert "Could not find any event" in err

    @mock.patch("wikitcms.event.ComposeEvent.__init__", return_value=None, autospec=True)
    @mock.patch("wikitcms.wiki.Wiki.get_current_event")
    def test_get_validation_event_guess_nomatch(self, fakecurr, fakeevent, fakemwp):
        """Check that when we guess but specify a release or milestone
        and the guessed release doesn't match the requested release or
        milestone, we fail.
        """
        event = wikitcms.event.ComposeEvent(self.site, 24, "Beta", "1.1")
        fakeevent.return_value = event
        with pytest.raises(ValueError) as err:
            self.site.get_validation_event(release=23)
            assert "does not match requested" in err
        with pytest.raises(ValueError) as err:
            self.site.get_validation_event(milestone="Final")
            assert "does not match requested" in err

    @mock.patch("fedfind.release.Compose.exists", return_value=True)
    @mock.patch("fedfind.release.Production.label", "RC-1.6")
    @mock.patch("fedfind.release.Production.cid", "Fedora-27-20171105.0")
    @mock.patch("fedfind.helpers.get_current_release", autospec=True, return_value=27)
    @mock.patch("mwclient.page.Page.__init__", return_value=None, autospec=True)
    @mock.patch("mwclient.page.Page.text", return_value=FAKE_CURRENT_COMPOSE)
    @mock.patch("wikitcms.page.NightlyPage", autospec=True)
    @mock.patch("wikitcms.page.ComposePage", autospec=True)
    def test_get_validation_page(
        self, fakecompose, fakenightly, faketext, fakeinit, fakegetcurr, fakecompexists
    ):
        """Various tests for get_validation_page."""
        # current event
        self.site.get_validation_page("Installation")
        fakecompose.assert_called_with(
            self.site, "24", "Installation", "Alpha", "1.1", dist="Fedora"
        )
        # old-school TC/RC
        self.site.get_validation_page("Installation", 23, "Final", "TC9")
        fakecompose.assert_called_with(self.site, 23, "Installation", "Final", "TC9", dist="Fedora")
        self.site.get_validation_page("Installation", 23, "Beta", "RC1")
        fakecompose.assert_called_with(self.site, 23, "Installation", "Beta", "RC1", dist="Fedora")
        # old-school nightly
        self.site.get_validation_page("Installation", 23, "Rawhide", "20151112")
        fakenightly.assert_called_with(
            self.site, 23, "Installation", "Rawhide", "20151112", dist="Fedora"
        )
        self.site.get_validation_page("Installation", 23, "Branched", "20151211")
        fakenightly.assert_called_with(
            self.site, 23, "Installation", "Branched", "20151211", dist="Fedora"
        )
        # Pungi 4 production/candidate
        self.site.get_validation_page("Installation", 24, "Alpha", "1.1")
        fakecompose.assert_called_with(self.site, 24, "Installation", "Alpha", "1.1", dist="Fedora")
        self.site.get_validation_page("Installation", 27, "Beta", "1.5", dist="Fedora-Modular")
        fakecompose.assert_called_with(
            self.site, 27, "Installation", "Beta", "1.5", dist="Fedora-Modular"
        )
        # Past 23, 'Final' milestone should be converted to 'RC'
        self.site.get_validation_page("Installation", 25, "Final", "1.1")
        fakecompose.assert_called_with(self.site, 25, "Installation", "RC", "1.1", dist="Fedora")
        # Pungi 4 nightly
        self.site.get_validation_page("Installation", 24, "Rawhide", "20160222.n.0")
        fakenightly.assert_called_with(
            self.site, 24, "Installation", "Rawhide", "20160222.n.0", dist="Fedora"
        )
        self.site.get_validation_page("Installation", 24, "Branched", "20160315.n.1")
        fakenightly.assert_called_with(
            self.site, 24, "Installation", "Branched", "20160315.n.1", dist="Fedora"
        )
        self.site.get_validation_page(
            "Installation", 27, "Branched", "20171110.n.1", dist="Fedora-Modular"
        )
        fakenightly.assert_called_with(
            self.site, 27, "Installation", "Branched", "20171110.n.1", dist="Fedora-Modular"
        )
        # Rawhide nightly compose ID
        self.site.get_validation_page("Installation", cid="Fedora-Rawhide-20180220.n.0")
        fakenightly.assert_called_with(
            self.site, "28", "Installation", "Rawhide", "20180220.n.0", dist="Fedora"
        )
        # Branched nightly compose ID
        self.site.get_validation_page("Installation", cid="Fedora-27-20171120.n.0")
        fakenightly.assert_called_with(
            self.site, "27", "Installation", "Branched", "20171120.n.0", dist="Fedora"
        )
        # Candidate compose ID
        self.site.get_validation_page("Installation", cid="Fedora-27-20171105.0")
        fakecompose.assert_called_with(self.site, "27", "Installation", "RC", "1.6", dist="Fedora")

        with pytest.raises(ValueError):
            # Non-nightly compose but no milestone
            self.site.get_validation_page("Installation", 24, "", "1.1")
        # Invalid composes
        with pytest.raises(ValueError):
            self.site.get_validation_page("Installation", 24, "Branched", "foobar")
        with pytest.raises(ValueError):
            self.site.get_validation_page("Installation", 24, "Branched", "TC1a")
        with pytest.raises(ValueError):
            self.site.get_validation_page("Installation", 24, "Branched", "1.1a")
        # looks kinda like a date but is not one
        with pytest.raises(ValueError):
            self.site.get_validation_page("Installation", 24, "Branched", "20161356")
        with pytest.raises(ValueError):
            self.site.get_validation_page("Installation", 24, "Branched", "20161356.n.0")
        with pytest.raises(ValueError):
            # invalid type
            self.site.get_validation_page("Installation", 24, "Branched", "20160314.x.0")

    def test_get_validation_page_guess_nightly(self):
        """Test get_validation_page guessing milestone for nightly
        events.
        """
        # fake_pginit1 will test the Rawhide path
        with mock.patch("wikitcms.page.NightlyPage.__init__", fake_pginit1):
            ret = self.site.get_validation_page("Installation", 24, "", "20160314.n.0")
        assert isinstance(ret, wikitcms.page.NightlyPage)
        assert ret.milestone == "Rawhide"
        # fake_pginit2 will test the Rawhide path
        with mock.patch("wikitcms.page.NightlyPage.__init__", fake_pginit2):
            ret = self.site.get_validation_page("Installation", 24, "", "20160314.n.0")
        assert isinstance(ret, wikitcms.page.NightlyPage)
        assert ret.milestone == "Branched"
        # ...and fake_pginit3 won't hit anywhere and should raise
        with mock.patch("wikitcms.page.NightlyPage.__init__", fake_pginit3):
            with pytest.raises(ValueError) as err:
                ret = self.site.get_validation_page("Installation", 24, "", "20160314.n.0")
                assert "Could not find any event" in err

    def test_get_validation_page_guess_nomatch(self, fakemwp):
        """Check that when we guess but specify a release or milestone
        and the guessed release doesn't match the requested release or
        milestone, we fail.
        """
        fakecurr = {
            "full": "32 Branched 20200407.n.0",
            "release": "32",
            "milestone": "Branched",
            "compose": "",
            "date": "20200407.n.0",
        }
        with mock.patch("wikitcms.wiki.Wiki.get_current_compose", return_value=fakecurr):
            with pytest.raises(ValueError) as err:
                self.site.get_validation_page("Installation", release=23)
                assert "does not match requested" in err
            with pytest.raises(ValueError) as err:
                self.site.get_validation_page("Installation", milestone="Final")
                assert "does not match requested" in err

    # we use the find_resultrow and ResultRow dummies from the top of
    # the file here
    @mock.patch("wikitcms.page.NightlyPage.__init__", return_value=None, autospec=True)
    @mock.patch("wikitcms.page.ComposePage.__init__", return_value=None, autospec=True)
    @mock.patch("wikitcms.page.NightlyPage.find_resultrow", fake_findrow)
    @mock.patch("wikitcms.page.ComposePage.find_resultrow", fake_findrow)
    @mock.patch("wikitcms.page.NightlyPage.add_results", autospec=True)
    @mock.patch("wikitcms.page.ComposePage.add_results", autospec=True)
    def test_report_validation_results(self, fakecompose, fakenightly, fakecompinit, fakenightinit):
        """Tests for report_validation_results."""
        restups = [
            wk.ResTuple(
                "Installation",
                "24",
                "Alpha",
                "1.1",
                "QA:Testcase_foo",
                status="pass",
                user="adamwill",
            ),
            wk.ResTuple(
                "Installation",
                "24",
                "Alpha",
                "1.1",
                "QA:Testcase_bar",
                status="pass",
                user="adamwill",
                section="testsec",
                env="testenv1",
            ),
            wk.ResTuple(
                "Installation",
                "24",
                "Alpha",
                "1.1",
                "QA:Testcase_bar",
                status="pass",
                user="adamwill",
                section="testsec",
                env="testenv2",
            ),
            wk.ResTuple(
                "Installation",
                "24",
                "Branched",
                "20160314.n.1",
                "QA:Testcase_foo",
                status="pass",
                user="adamwill",
            ),
            wk.ResTuple(
                "Installation",
                "24",
                "Branched",
                "20160314.n.1",
                "QA:Testcase_bar",
                status="pass",
                user="adamwill",
                section="testsec",
                env="testenv1",
            ),
            wk.ResTuple(
                "Installation",
                "24",
                "Branched",
                "20160314.n.1",
                "QA:Testcase_bar",
                status="pass",
                user="adamwill",
                section="testsec",
                env="testenv2",
            ),
        ]
        self.site.report_validation_results(restups)
        # this ought to call add_results once for each page.
        for fake in (fakecompose, fakenightly):
            assert len(fake.call_args_list) == 1
            compargs = fake.call_args
            # compargs[0] is all args as a list, add_results takes a
            # single arg which is the resdict first item is self (when
            # autospec is used)
            resdict = compargs[0][1]
            # resdict should have exactly two entries, one per test
            # instance
            assert len(resdict) == 2
            # we're going to sort the dict items for ease of testing.
            items = sorted(list(resdict.items()), key=lambda x: x[0].name, reverse=True)
            # first resdict entry: should be for the 'foo' test case...
            (key, value) = items[0]
            assert key.name == "QA:Testcase_foo.."
            # ...and value should be a 1-item list of a single 2-tuple,
            # env '', result a Result instance
            assert len(value) == 1
            (env, res) = value[0]
            assert env == ""
            assert isinstance(res, wikitcms.result.Result)
            (key, value) = items[1]
            # second resdict entry: should be for the 'bar' test case
            # with section name...
            assert key.name == "QA:Testcase_bar.testsec."
            # ... and value should be a 2-item list of 2-tuples,
            # differing in the environment
            assert len(value) == 2
            (env1, res) = value[0]
            assert env1 == "testenv1"
            assert isinstance(res, wikitcms.result.Result)
            (env1, res) = value[1]
            assert env1 == "testenv2"
            assert isinstance(res, wikitcms.result.Result)

    @mock.patch("mwclient.listing.Category.members")
    def test_testtypes(self, fakemembers, fakemwp):
        """Tests for testtypes and, incidentally, matrices."""
        # this is more or less what the return of
        # category.members(generator=False) looks like - it's an
        # mwclient 'List' instance
        instmx = {"title": "Template:Installation test matrix", "pageid": 54437, "ns": 10}
        basemx = {"title": "Template:Base test matrix", "pageid": 54433, "ns": 10}
        fakemembers.return_value = [instmx, basemx]
        assert self.site.testtypes == ["Installation", "Base"]

    @mock.patch("mwclient.listing.Category.members")
    def test_modular_testtypes(self, fakemembers, fakemwp):
        """Tests for modular testtypes and, incidentally, matrices."""
        # this is more or less what the return of
        # category.members(generator=False) looks like - it's an
        # mwclient 'List' instance
        instmx = {"title": "Template:Installation Modular test matrix", "pageid": 77036, "ns": 10}
        servmx = {"title": "Template:Server Modular test matrix", "pageid": 77040, "ns": 10}
        fakemembers.return_value = [instmx, servmx]
        assert self.site.modular_testtypes == ["Installation", "Server"]

    @mock.patch("mwclient.Site.login", autospec=True)
    @mock.patch("mwclient.Site.site_init", autospec=True)
    def test_login(self, fakeinit, fakelogin):
        """Tests for login(), including all the OpenID stuff."""
        # self.site should just wrap mwclient as it's not fp.o
        self.site.login(username="foo", password="bar")
        assert fakelogin.call_count == 1
        assert fakelogin.call_args[1]["username"] == "foo"
        assert fakelogin.call_args[1]["password"] == "bar"
        # check host splitting
        fakelogin.reset_mock()
        site = wk.Wiki(["https", "fjajah"], do_init=False, force_login=False)
        site.login(username="foo", password="bar")
        assert fakelogin.call_count == 1

        # now, activate the openid bits
        fakelogin.reset_mock()
        site = wk.Wiki("fedoraproject.org", do_init=False, force_login=False)
        # have to fake this up
        site.version = (1, 30)
        site.login()
        assert isinstance(site.connection.auth, openidc_client.requestsauth.OpenIDCClientAuther)
        assert site.connection.auth.scopes == ["openid", "https://fedoraproject.org/wiki/api"]
        assert isinstance(site.connection.auth.client, openidc_client.OpenIDCClient)
        assert site.connection.auth.client.app_id == "wikitcms"
        assert site.connection.auth.client.idp == "https://id.fedoraproject.org/openidc/"
        assert site.connection.auth.client.idp_mapping == {
            "Token": "Token",
            "Authorization": "Authorization",
        }
        assert site.connection.auth.client.client_id == "wikitcms"
        # <hacker voice>I'M IN</hacker voice>
        assert site.connection.auth.client.client_secret == "notsecret"
        assert site.connection.auth.client.useragent == "wikitcms"
        # this is when we can't import the OpenID bits - the idea is
        # we still work for non-OpenID paths in that case and only die
        # here
        with mock.patch("wikitcms.wiki.OpenIDCClient", None):
            with pytest.raises(ImportError):
                site.login()
        with mock.patch("wikitcms.wiki.OpenIDCClientAuther", None):
            with pytest.raises(ImportError):
                site.login()


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""The Wiki class here extends mwclient's Site class with additional
Wikitcms-specific functionality, and convenience features like stored
user credentials.
"""

from collections import namedtuple
import datetime
import re

import fedfind.helpers
import mwclient
from productmd.composeinfo import get_date_type_respin

import wikitcms.event
from wikitcms.exceptions import NoPageError, NotFoundError, TooManyError
import wikitcms.helpers
import wikitcms.listing
import wikitcms.page
import wikitcms.result

try:
    from openidc_client import OpenIDCClient
except ImportError:
    OpenIDCClient = None
try:
    from openidc_client.requestsauth import OpenIDCClientAuther
except ImportError:
    OpenIDCClientAuther = None


def _check_compose(compose):
    """Trivial checker shared between get_validation_event and
    get_validation_compose.
    """
    date = fedfind.helpers.date_check(compose, out="obj", fail_raise=False)
    # all nightlies after F24 branch are Pungi 4-style; plain date
    # is never a valid 'compose' value after that date, 2016-02-23
    if date and date < datetime.datetime(2016, 2, 23):
        return "date"

    # check if we have a valid Pungi4-style identifier
    try:
        (date, typ, respin) = get_date_type_respin(compose)
        if date and typ and respin is not None:
            if fedfind.helpers.date_check(date, fail_raise=False):
                return "date"
    except ValueError:
        pass

    # regex to match TC/RC names: TC1, RC10, RC23.6
    patt = re.compile(r"^[TR]C\d+\.?\d*$")
    if patt.match(compose.upper()):
        return "compose"

    # regex for Pungi 4 milestone composes: 1.1, 1.2 ... 10.10 ...
    patt = re.compile(r"^\d+\.\d+$")
    if patt.match(compose):
        return "compose"

    # if nothing matched, value is invalid
    raise ValueError(
        "Compose must be a TC/RC identifier (TC1, RC3...) for pre-"
        "Fedora 24 milestone composes, a Pungi 4 milestone compose"
        " identifier (1.1, 10.10...) for post-Fedora 23 milestone "
        "composes, a date in YYYYMMDD format (for pre-Fedora 24 "
        "nightlies) or a Pungi 4 nightly identifier (20160308.n.0"
        ", 20160310.n.2) for post-Fedora 23 nightlies."
    )


# I'd really like to use namedlist.namedtuple, but it isn't widely
# available.
class ResTuple(
    namedtuple(
        "ResTuple",
        "testtype release milestone compose "
        "testcase section testname env status user bugs "
        "comment bot cid dist",
    )
):
    """namedtuple (with default values) used for report_validation_
    results(). See that method's docstring for details.
    """

    # pylint: disable=too-many-arguments, too-many-locals
    def __new__(
        cls,
        testtype,
        release="",
        milestone="",
        compose="",
        testcase="",
        section="",
        testname="",
        env="",
        status="",
        user="",
        bugs="",
        comment="",
        bot=False,
        cid="",
        dist="Fedora",
    ):
        return super(ResTuple, cls).__new__(
            cls,
            testtype,
            release,
            milestone,
            compose,
            testcase,
            section,
            testname,
            env,
            status,
            user,
            bugs,
            comment,
            bot,
            cid,
            dist,
        )


class Wiki(mwclient.Site):
    """Extends the mwclient.Site class with some extra capabilities."""

    # parent class has a whole bunch of args, so just pass whatever
    # through. always init this the same as a regular mwclient.Site
    # instance. we have to disable this warning as the args must be
    # in this exact order to maintain compatibility
    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, host="fedoraproject.org", *args, **kwargs):
        super(Wiki, self).__init__(host, *args, **kwargs)
        # override the 'pages' property so it returns wikitcms Pages when
        # appropriate
        self.pages = wikitcms.listing.TcmsPageList(self)

    @property
    def current_compose(self):
        """A dict of the key / value pairs from the CurrentFedora
        Compose page which is the canonical definition of the 'current'
        primary arch validation testing compose. You can usually expect
        keys full, release, date, milestone, and compose. The page is
        normally written by ValidationEvent.update_current().
        """
        return self.get_current_compose(dist="Fedora")

    @property
    def current_event(self):
        """The current event, as a ValidationEvent instance. Will be a
        ComposeEvent or a NightlyEvent."""
        return self.get_current_event(dist="Fedora")

    @property
    def current_modular_compose(self):
        """A dict of the key / value pairs from the CurrentFedora
        ModularCompose page which is the canonical definition of the
        'current' modular primary arch validation testing compose. You
        can usually expect keys full, release, date, milestone, and
        compose. The page is normally written by
        ValidationEvent.update_current().
        """
        return self.get_current_compose(dist="Fedora-Modular")

    @property
    def current_modular_event(self):
        """The current modular event, as a ValidationEvent instance.
        Will be a ComposeEvent or a NightlyEvent."""
        return self.get_current_event(dist="Fedora-Modular")

    @property
    def matrices(self):
        """A list of dicts representing pages in the test matrix
        template category. These are the canonical definition of
        'existing' test types. Done this way - rather than using the
        category object's ability to act as an iterator over its member
        page objects - because this method respects the sort order of
        the member pages, whereas the other does not. The sort order is
        used in creating the overview summary page.
        """
        return self.get_matrices(dist="Fedora")

    @property
    def testtypes(self):
        """Test types, derived from the matrix page names according to
        a naming convention. A list of strings.
        """
        return self.get_testtypes(dist="Fedora")

    @property
    def modular_matrices(self):
        """A list of dicts representing pages in the modular test
        matrix template category. These are the canonical definition of
        'existing' test types. Done this way - rather than using the
        category object's ability to act as an iterator over its member
        page objects - because this method respects the sort order of
        the member pages, whereas the other does not. The sort order is
        used in creating the overview summary page.
        """
        return self.get_matrices(dist="Fedora-Modular")

    @property
    def modular_testtypes(self):
        """Test types, derived from the matrix page names according to
        a naming convention. A list of strings.
        """
        return self.get_testtypes(dist="Fedora-Modular")

    def get_current_compose(self, dist="Fedora"):
        """A dict of the key / value pairs from the Current{dist}
        Compose page which is the canonical definition of the 'current'
        primary arch validation testing compose. You can usually expect
        keys full, release, date, milestone, and compose. The page is
        normally written by ValidationEvent.update_current().
        """
        wikitcms.helpers.validate_dist(dist)
        currdict = dict()
        valpatt = re.compile(r"^\| *?(\w+?) *?= *([\w .]*?) *$", re.M)
        page = self.pages[f"Template:Current{dist}Compose"]
        for match in valpatt.finditer(page.text()):
            currdict[match.group(1)] = match.group(2)
        return currdict

    def get_current_event(self, dist="Fedora"):
        """Returns the current event for the given dist as a
        ValidationEvent instance. May return None if there is no
        current event for the given dist (e.g. none has yet been
        created for that dist).
        """
        wikitcms.helpers.validate_dist(dist)
        curr = self.get_current_compose(dist=dist)
        if not curr:
            return None
        # Use of 'max' plus get_validation_event handles getting us
        # the right kind of event.
        return self.get_validation_event(
            release=curr["release"],
            milestone=curr["milestone"],
            compose=max(curr["date"], curr["compose"]),
            dist=dist,
        )

    def get_matrices(self, dist="Fedora"):
        """A list of dicts representing pages in the test matrix
        template category for the specified dist. These are the
        canonical definition of 'existing' test types. Done this way -
        rather than using the category object's ability to act as an
        iterator over its member page objects - because this method
        respects the sort order of the member pages, whereas the other
        does not. The sort order is used in creating the overview
        summary page.
        """
        wikitcms.helpers.validate_dist(dist)
        # dist names have a specific form, we expect a category derived
        # from the name to exist. If dist is "Fedora", we know the name
        if dist == "Fedora":
            catname = "Category:QA test matrix templates"
        else:
            # drop "Fedora-"
            dstring = dist[7:]
            catname = f"Category:QA {dstring} test matrix templates"
        category = self.pages[catname]
        return category.members(generator=False)

    def get_testtypes(self, dist="Fedora"):
        """Returns test types (as a list of strings) for a given dist,
        derived from the matrix page names for that dist.
        """
        wikitcms.helpers.validate_dist(dist)
        # drop "Fedora-"
        dstring = dist[7:]
        if dstring:
            dstring = f"{dstring} "
        return [
            m["title"].replace("Template:", "").replace(f" {dstring}test matrix", "")
            for m in self.get_matrices(dist=dist)
        ]

    def login(self, *args, **kwargs):  # pylint: disable=signature-differs
        """Login method, overridden to use openidc auth when necessary.
        This will open a browser window and run through FAS auth on
        the first use, then a token will be saved that will allow auth
        for a while, when the token expires, the web auth process will
        pop up again.
        """
        use_openidc = False
        # This probably breaks on private wikis, but wikitcms isn't
        # ever used with any of those, AFAIK.
        host = self.host
        if isinstance(host, (list, tuple)):
            host = host[1]
        if host.endswith("fedoraproject.org") and self.version[:2] >= (1, 29):
            use_openidc = True
        if not use_openidc:
            # just work like mwclient
            super(Wiki, self).login(*args, **kwargs)

        # Fedora wiki since upgrade to 1.29 doesn't allow native
        # mediawiki auth with FAS creds any more, it only allows
        # auth via OpenID Connect. For this case we set up an
        # openidc auther, call site_init() to trigger auth and
        # update the site properties, and return.
        if OpenIDCClient is None:
            raise ImportError("python-openidc-client is needed for OIDC")
        if OpenIDCClientAuther is None:
            raise ImportError("python-openidc-client 0.4.0 or higher is " "required for OIDC")
        client = OpenIDCClient(
            app_identifier="wikitcms",
            id_provider=f"https://id.{host}/openidc/",
            id_provider_mapping={"Token": "Token", "Authorization": "Authorization"},
            client_id="wikitcms",
            client_secret="notsecret",
            useragent="wikitcms",
        )

        auther = OpenIDCClientAuther(client, ["openid", "https://fedoraproject.org/wiki/api"])

        self.connection.auth = auther
        self.site_init()

    def add_to_category(self, page_name, category_name, summary=""):
        """Add a given page to a given category if it is not already a
        member. Takes strings for the names of the page and the
        category, not mwclient objects.
        """
        page = self.pages[page_name]
        text = page.text()
        if category_name not in text:
            text += f"\n[[{category_name}]]"
            page.save(text, summary, createonly=False)

    def walk_category(self, category):
        """Simple recursive category walk. Returns a list of page
        objects that are members of the parent category or its
        sub-categories, to any level of recursion. 14 is the Category:
        namespace.
        """
        pages = dict()
        for page in category:
            if page.namespace == 14:
                sub_pages = self.walk_category(page)
                for sub_page in sub_pages:
                    pages[sub_page.name] = sub_page
            else:
                pages[page.name] = page
        pages = pages.values()
        return pages

    def allresults(self, prefix=None, start=None, redirects="all", end=None):
        """A generator for pages in the Test Results: namespace,
        similar to mwclient's allpages, allcategories etc. generators.
        This is a TcmsPageList, so it returns wikitcms objects when
        appropriate. Note, if passing prefix, start or end, leave out
        the "Test Results:" part of the name.
        """
        gen = wikitcms.listing.TcmsPageList(
            self, prefix=prefix, start=start, namespace=116, redirects=redirects, end=end
        )
        return gen

    def alltestdays(self, prefix=None, start=None, redirects="all", end=None):
        """A generator for pages in the Test Day: namespace,
        similar to mwclient's allpages, allcategories etc. generators.
        This is a TcmsPageList, so it returns wikitcms objects when
        appropriate. Note, if passing prefix, start or end, leave out
        the "Test Day:" part of the name.
        """
        gen = wikitcms.listing.TcmsPageList(
            self, prefix=prefix, start=start, namespace=114, redirects=redirects, end=end
        )
        return gen

    def get_validation_event(self, release="", milestone="", compose="", cid="", dist="Fedora"):
        # we could factor this out but I think it's fine
        # pylint: disable=too-many-arguments, too-many-branches
        """Get an appropriate ValidationEvent object for the values
        given. As with get_validation_page(), this method is for
        sloppy instantiation of pages that follow the rules. This
        method has no required arguments and tries to figure out
        what you want from what you give it. It will raise errors
        if what you give it is impossible to interpret or if it
        tries and comes up with an inconsistent-seeming result.

        dist specifies a 'dist' (compose shortname) other than the
        default "Fedora" to find an event for. If you pass a compose
        ID as cid, any value you pass for dist will be ignored; we'll
        instead parse the dist value out of the compose ID.

        If you pass a numeric release, a milestone, and a valid
        compose (TC/RC or date), it will give you the appropriate
        event, whether it exists or not. All it really does in this
        case is pick NightlyEvent or ComposeEvent for you. If you
        don't fulfill any of those conditions, it'll need to do
        some guessing/assumptions, and in some of those cases it
        will only return an Event *that actually exists*, and may
        raise exceptions if you passed a particularly pathological
        set of values.

        If you don't pass a compose argument it will get the current
        event; if you passed either of the other arguments and they
        don't match the current event, it will raise an error. It
        follows that calling this with no arguments just gives you
        current_event.

        If you pass a date as compose with no milestone, it will see
        if there's a Rawhide nightly and return it if so, otherwise it
        will see if there's a Branched nightly and return that if so,
        otherwise raise an error. It follows that you can't get the
        page for an event that *doesn't exist yet* this way: you must
        instantiate it directly or call this method with a milestone.

        It will not attempt to guess a milestone for TC/RC composes;
        it will raise an exception in this case.

        The guessing bits require wiki roundtrips, so they will be
        slower than instantiating a class directly or using this
        method with sufficient information to avoid guessing.
        """
        if cid:
            (dist, release, milestone, compose) = wikitcms.helpers.cid_to_event(cid)
        if not compose or not release:
            # Can't really make an educated guess without a compose
            # and release, so just get the current event and return it
            # if it matches any other values passed.
            event = self.get_current_event(dist=dist)
            if release and event.release != release:
                raise ValueError(
                    f"get_validation_event(): Guessed event release {event.release} does "
                    f"not match requested release {release}"
                )
            if milestone and event.milestone != milestone:
                raise ValueError(
                    f"get_validation_event(): Guessed event milestone {event.milestone} "
                    "does not match specified milestone {milestone}"
                )
            # all checks OK
            return event

        if _check_compose(compose) == "date":
            if milestone:
                return wikitcms.event.NightlyEvent(
                    self, release=release, milestone=milestone, compose=compose, dist=dist
                )

            # we have a date and no milestone. Try both and return
            # whichever exists. We check whether the first result
            # page has any contents so that if someone mistakenly
            # creates the wrong event, we can clean up by blanking
            # the pages, rather than by getting an admin to
            # actually *delete* them.
            rawev = wikitcms.event.NightlyEvent(self, release, "Rawhide", compose, dist=dist)
            pgs = rawev.result_pages
            if pgs and pgs[0].text():
                return rawev
            brev = wikitcms.event.NightlyEvent(self, release, "Branched", compose, dist=dist)
            pgs = brev.result_pages
            if pgs and pgs[0].text():
                return brev
            # Here, we failed to guess. Boohoo.
            raise ValueError(
                "get_validation_event(): Could not find any event for "
                f"release {release} and date {compose}."
            )

        if _check_compose(compose) == "compose":
            compose = str(compose).upper()
            if not milestone:
                raise ValueError(
                    "get_validation_event(): For a TC/RC compose, a milestone "
                    "- Alpha, Beta, or Final - must be specified."
                )
            # With Pungi 4, the 'Final' milestone became 'RC', let's
            # be nice and convert it
            if int(release) > 23 and milestone.lower() == "final":
                milestone = "RC"
            return wikitcms.event.ComposeEvent(
                self, release, milestone, compose, dist=dist, cid=cid
            )

        # We should never get here, but just in case.
        raise ValueError("get_validation_event(): Something very strange happened.")

    def get_validation_page(
        self, testtype, release="", milestone="", compose="", cid="", dist="Fedora"
    ):
        # we could factor this out but I think it's fine
        # pylint: disable=too-many-arguments
        """Get an appropriate ValidationPage object for the values
        given. As with get_validation_event(), this method is for
        sloppy instantiation of pages that follow the rules. This
        method has no required arguments except the testtype and tries
        to figure out what you want from what you give it. It will
        raise errors if what you give it is impossible to interpret or
        if it tries and comes up with an inconsistent-seeming result.

        dist specifies a 'dist' (compose shortname) other than the
        default "Fedora" to find a page for. If you pass a compose ID
        as cid, any value you pass for dist will be ignored; we'll
        instead parse the dist value out of the compose ID.

        If you pass a numeric release, a milestone, and a valid
        compose (TC/RC or date), it will give you the appropriate
        event, whether it exists or not. All it really does in this
        case is pick NightlyEvent or ComposeEvent for you. If you
        don't fulfill any of those conditions, it'll need to do
        some guessing/assumptions, and in some of those cases it
        will only return an Event *that actually exists*, and may
        raise exceptions if you passed a particularly pathological
        set of values.

        If you don't pass a compose argument it will get the page for
        the current event; if you passed either of the other
        arguments and they don't match the current event, it will
        raise an error. It follows that calling this with no arguments
        just gives you the page of the specified test type for the
        current event.

        If you pass a date as compose with no milestone, it will see
        if there's a Rawhide nightly and return it if so, otherwise it
        will see if there's a Branched nightly and return that if so,
        otherwise raise an error. It follows that you can't get the
        page for an event that *doesn't exist yet* this way: you must
        instantiate it directly or call this method with a milestone.

        It will not attempt to guess a milestone for TC/RC composes;
        it will raise an exception in this case.

        The guessing bits require wiki roundtrips, so they will be
        slower than instantiating a class directly or using this
        method with sufficient information to avoid guessing.
        """
        if cid:
            (dist, release, milestone, compose) = wikitcms.helpers.cid_to_event(cid)
        if not compose or not release:
            # Can't really make an educated guess without a compose
            # and release, so just get the current event and return it
            # if it matches any other values passed.
            curr = self.get_current_compose(dist=dist)
            page = self.get_validation_page(
                testtype,
                release=curr["release"],
                milestone=curr["milestone"],
                compose=max(curr["compose"], curr["date"]),
                dist=dist,
            )
            if release and page.release != release:
                raise ValueError(
                    f"get_validation_page(): Guessed page release {page.release} does "
                    f"not match requested release {release}"
                )
            if milestone and page.milestone != milestone:
                raise ValueError(
                    f"get_validation_page(): Guessed page milestone {page.milestone} "
                    f"does not match specified milestone {milestone}"
                )
            return page

        if _check_compose(compose) == "date":
            if milestone:
                return wikitcms.page.NightlyPage(
                    self, release, testtype, milestone, compose, dist=dist
                )

            # date, but no milestone
            rawpg = wikitcms.page.NightlyPage(
                self, release, testtype, "Rawhide", compose, dist=dist
            )
            if rawpg.exists:
                return rawpg
            brpg = wikitcms.page.NightlyPage(
                self, release, testtype, "Branched", compose, dist=dist
            )
            if brpg.exists:
                return brpg
            # Here, we failed to guess. Boohoo.
            raise ValueError(
                "get_validation_page(): Could not find any event for "
                f"release {release} and date {compose}."
            )

        if _check_compose(compose) == "compose":
            if not milestone:
                raise ValueError(
                    "get_validation_page(): For a milestone compose, a "
                    " milestone - Alpha, Beta, or Final - must be specified."
                )
            # With Pungi 4, the 'Final' milestone became 'RC', let's
            # be nice and convert it
            if int(release) > 23 and milestone.lower() == "final":
                milestone = "RC"
            return wikitcms.page.ComposePage(self, release, testtype, milestone, compose, dist=dist)

        # We should never get here, but just in case.
        raise ValueError("get_validation_page(): Something very strange happened.")

    def report_validation_results(self, reslist, allowdupe=False):
        # pylint: disable=too-many-locals,too-many-branches
        """High-level result reporting function. Pass it an iterable
        of objects identifying results. It's pretty forgiving about
        what these can be. They can be any kind of sequence or
        iterable containing up to 15 values in the following order:
        (testtype, release, milestone, compose, testcase, section,
        testname, env, status, user, bugs, comment, bot, cid, dist).
        They can also be any mapping type (e.g. a dict) with enough of
        the 15 keys set (see below for requirements).

        You may find it convenient to import the ResTuple class from
        this module and pass your results as instances of it: it is
        a namedtuple with default values which uses the names given
        above, so you can avoid having to pad values you don't need to
        set and conveniently read back items from the tuple after
        you've created it, if you like.

        Any value of the result item can be absent, or set to
        something empty or falsey. However, to be successfully
        reported, each item must meet these conditions:

        * 'testtype', 'release', 'milestone', 'compose', 'dist' and
        'cid' must identify a single validation page
        using get_validation_page() (at least 'testtype' must always
        be set).
        * 'testcase', 'section' and 'testname' must identify a single
        test instance using ValidationPage.find_resultrow()
        * 'env' must uniquely identify one of the test instance's
        'environments' (the columns into which results can be entered;
        it can be empty if there is only one for this test instance)
        *'status' must indicate the result status.

        'user', 'bugs', and 'comment' can always be left empty if
        desired; if 'user' is not specified, the mediawiki username
        (lower-cased) will be used.

        All values should be strings, except 'bugs', which should be
        an iterable of digit strings (though iterable of ints will be
        tolerated), and 'bot' which should be True if the result is
        from some sort of automated testing system (not a human).

        Returns a 2-tuple of lists of objects indicating any failures.
        The first list is of submissions that failed because there was
        insufficient identifying data. The second list is of
        submissions that failed because they were 'dupes' - the given
        user has already reported a result for the given test and env.
        If allowdupe is True, duplicate reports will be allowed and
        this list will always be empty.

        The items in each list have the same basic layout as the input
        results items. The 'insufficients' list will contain the exact
        same objects that were provided as input. The 'duplicates'
        list is reconstructed and will often contain more or corrected
        data compared to the input tuples. Its members are *always*
        instances of the ResTuple namedtuple class, which provides
        access to the fields by the names given above.

        Uses get_validation_page() to guess the desired page and then
        constructs a result dict to pass to the ValidationPage
        add_results() method.
        """
        pagedict = dict()  # KEY: (testtype, release, milestone, compose, cid, dist)
        insufficients = list()
        dupes = list()

        for resitem in reslist:
            # Convert the item into a ResTuple.
            try:
                restup = ResTuple(**resitem)
            except TypeError:
                restup = ResTuple(*resitem)
            if not restup.status:
                # It doesn't make sense to allow {{result|none}}
                # from this high-level function.
                insufficients.append(resitem)
                continue
            user = restup.user
            if not user:
                # If no username was given, guess at the wiki account
                # name, lower-cased.
                user = self.username.lower()
            key = (
                restup.testtype,
                restup.release,
                restup.milestone,
                restup.compose,
                restup.cid,
                restup.dist,
            )
            # We construct a dict to sort the results by page. The
            # value for each page is a 2-tuple containing the actual
            # ValidationPage object and the 'results dictionary' we
            # will pass to page.add_results() once we have iterated
            # through the full result list (unless the page cannot be
            # found, in which case the value is None.)
            if key not in pagedict:
                # We haven't yet encountered this page, so we'll try
                # and find it and add it to the dict.
                pagedict[key] = dict()
                try:
                    pagedict[key] = (self.get_validation_page(*key), dict())
                except ValueError:
                    # This means we couldn't find a page from the info
                    # provided; set the pagedict entry for this key to
                    # (None, None) to cache that information, add this
                    # result to the 'insufficients' list, and move on
                    # to the next.
                    pagedict[key] = (None, None)
                    insufficients.append(resitem)
                    continue
            elif pagedict[key] == (None, None):
                # This is when we've already tried to find a page from
                # the same (testtype, release, milestone, compose) and
                # come up empty, so append the restup to the insuffs
                # list and move on.
                insufficients.append(resitem)
                continue

            # The code from here on gets hit whenever we can find the
            # correct page for a restup. We need to find the correct
            # ResultRow from the testcase, section, testname and
            # env, and produce a Result from status, username, bugs,
            # and comment. The keys in resdict are ResultRow
            # instances, so we check if there's already one that's
            # 'equal' to the ResultRow for this restup. If so, we
            # append the (env, Result) tuple to the list that is the
            # value for that key. If not, we add a new entry to the
            # resdict. It's vital that the resdict contain only *one*
            # entry for any given ResultRow, or else the edit will go
            # squiffy.
            (page, resdict) = pagedict[key]
            try:
                myrow = page.find_resultrow(
                    restup.testcase, restup.section, restup.testname, restup.env
                )
                myres = wikitcms.result.Result(
                    restup.status, user, restup.bugs, restup.comment, restup.bot
                )
            except (NoPageError, NotFoundError, TooManyError):
                # We couldn't find precisely one result row from the
                # provided information.
                insufficients.append(resitem)
                continue

            done = False
            for row in resdict.keys():
                if myrow.matches(row):
                    resdict[row].append((restup.env, myres))
                    done = True
                    break
            if not done:
                resdict[myrow] = [(restup.env, myres)]

        # Finally, we've sorted our restups into the right shape. Now
        # all we do is throw each page's resdict at its add_results()
        # method and grab the return value, which is a list of tuples
        # representing results that were 'dupes'. We then reconstruct
        # a ResTuple for each dupe, and return the lists of insuffs
        # and dupes.
        for (page, resdict) in pagedict.values():
            if page:
                _dupes = page.add_results(resdict, allowdupe)
                for (row, env, result) in _dupes:
                    dupes.append(
                        ResTuple(
                            page.testtype,
                            page.release,
                            page.milestone,
                            page.compose,
                            row.testcase,
                            row.section,
                            row.name,
                            env,
                            result.status,
                            result.user,
                            result.bugs,
                            result.comment,
                            result.bot,
                            "",
                            page.dist,
                        )
                    )

        return (insufficients, dupes)


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2016 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument, too-many-arguments
# pylint: disable=invalid-name

"""Tests for event.py."""

from unittest import mock

import fedfind.release
import mwclient.errors
import mwclient.page
import pytest

import wikitcms.event
import wikitcms.listing
import wikitcms.wiki as wk
from wikitcms.exceptions import FedfindNotFoundError


class TestEventFedfind:
    """Tests related to fedfind release discovery from validation
    events.
    """

    # the 'fjajah' is just to make sure we're running offline; if I
    # screw up and add a test that hits the network it'll cause the
    # tests to hang/fail instead of succeeding a bit slower
    site = wk.Wiki("fjajah", do_init=False, force_login=False)

    @mock.patch("fedfind.release.Compose.exists", True)
    @mock.patch("fedfind.release.Compose.all_images", ["foo"])
    def test_candidate_ff_release_compose(self):
        """Straightforward ff_release test case for a candidate
        compose which is properly synced to stage. Should work the
        same whether or not we pass the cid hint, both properties
        should exist and be the Compose instance.
        """
        event = wikitcms.event.ComposeEvent(
            self.site, "27", "RC", "1.6", cid="Fedora-27-20171105.0"
        )
        assert isinstance(event.ff_release, fedfind.release.Compose)
        assert isinstance(event.ff_release_images, fedfind.release.Compose)
        event = wikitcms.event.ComposeEvent(self.site, "27", "RC", "1.6")
        assert isinstance(event.ff_release, fedfind.release.Compose)
        assert isinstance(event.ff_release_images, fedfind.release.Compose)

    @mock.patch("fedfind.helpers.cid_from_label", return_value="")
    @mock.patch("fedfind.release.Compose.exists", False)
    @mock.patch("fedfind.release.Production.all_images", ["foo"])
    @mock.patch("fedfind.release.Production.cid", "Fedora-27-20171105.0")
    def test_candidate_ff_release_compose_gap(self, fakecidfromlabel):
        """Test the 'compose gap' case: this occurs when a candidate
        compose has just been created, but not yet synced to stage,
        and also has not yet appeared in PDC. In this case, without
        the 'cid' hint, we will not be able to find the fedfind
        release associated with the event. With the hint, we should
        find it in the non-synced location (as a fedfind Production).
        """
        event = wikitcms.event.ComposeEvent(
            self.site, "27", "RC", "1.6", cid="Fedora-27-20171105.0"
        )
        assert isinstance(event.ff_release, fedfind.release.Production)
        assert isinstance(event.ff_release_images, fedfind.release.Production)
        event = wikitcms.event.ComposeEvent(self.site, "27", "RC", "1.6")
        with pytest.raises(FedfindNotFoundError):
            print(event.ff_release)
        with pytest.raises(FedfindNotFoundError):
            print(event.ff_release_images)

    @mock.patch("fedfind.helpers.cid_from_label", return_value="Fedora-27-20171105.0")
    @mock.patch("fedfind.release.Compose.exists", False)
    @mock.patch("fedfind.release.Production.all_images", ["foo"])
    @mock.patch("fedfind.release.Production.cid", "Fedora-27-20171105.0")
    def test_candidate_ff_release_compose_gap_pdc(self, fakecidfromlabel):
        """Test the case where the candidate compose has not yet
        synced to stage, but has appeared in PDC. In this case, we
        should find the ff_release in the non-synced location (as a
        fedfind Production) with or without the cid hint.
        """
        event = wikitcms.event.ComposeEvent(
            self.site, "27", "RC", "1.6", cid="Fedora-27-20171105.0"
        )
        assert isinstance(event.ff_release, fedfind.release.Production)
        assert isinstance(event.ff_release_images, fedfind.release.Production)
        event = wikitcms.event.ComposeEvent(self.site, "27", "RC", "1.6")
        assert isinstance(event.ff_release, fedfind.release.Production)
        assert isinstance(event.ff_release_images, fedfind.release.Production)

    @mock.patch("fedfind.release.Compose.exists", True)
    @mock.patch("fedfind.release.Compose.all_images", [])
    @mock.patch("fedfind.release.Production.all_images", ["foo"])
    @mock.patch("fedfind.release.Production.cid", "Fedora-27-20171105.0")
    def test_candidate_ff_release_compose_exists_no_images(self):
        """Test a potential tricky case where a candidate compose
        tree exists on stage but the images haven't shown up in it
        yet. With the cid hint, the event's ff_release should be the
        Compose instance, but its ff_release_images should be the
        Production instance. Without the hint, we won't get images.
        """
        event = wikitcms.event.ComposeEvent(
            self.site, "27", "RC", "1.6", cid="Fedora-27-20171105.0"
        )
        assert isinstance(event.ff_release, fedfind.release.Compose)
        assert isinstance(event.ff_release_images, fedfind.release.Production)
        event = wikitcms.event.ComposeEvent(self.site, "27", "RC", "1.6")
        assert isinstance(event.ff_release, fedfind.release.Compose)
        with pytest.raises(FedfindNotFoundError):
            assert event.ff_release_images

    @mock.patch("fedfind.release.BranchedNightly.exists", True)
    @mock.patch("fedfind.release.BranchedNightly.all_images", ["foo"])
    def test_candidate_ff_release_nightly(self):
        """Straightforward ff_release test case for a nightly
        compose which exists and has images.
        """
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        assert isinstance(event.ff_release, fedfind.release.BranchedNightly)
        assert isinstance(event.ff_release_images, fedfind.release.BranchedNightly)

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_candidate_ff_release_nightly_no_images(self):
        """ff_release test case for a nightly compose which doesn't
        exist and has no images. We get ff_release (as fedfind doesn't
        do an existence check in this case), but not images.
        """
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        assert isinstance(event.ff_release, fedfind.release.BranchedNightly)
        with pytest.raises(FedfindNotFoundError):
            assert event.ff_release_images

    @mock.patch("fedfind.release.ModularCompose.exists", True)
    @mock.patch("fedfind.release.ModularCompose.all_images", ["foo"])
    def test_candidate_ff_release_modular(self):
        """Straightforward ff_release test case for a modular
        candidate compose which is properly synced to stage.
        """
        event = wikitcms.event.ComposeEvent(
            self.site,
            "27",
            "Beta",
            "1.5",
            dist="Fedora-Modular",
            cid="Fedora-Modular-27-20171108.2",
        )
        assert isinstance(event.ff_release, fedfind.release.ModularCompose)
        assert isinstance(event.ff_release_images, fedfind.release.ModularCompose)
        event = wikitcms.event.ComposeEvent(self.site, "27", "Beta", "1.5", dist="Fedora-Modular")
        assert isinstance(event.ff_release, fedfind.release.ModularCompose)
        assert isinstance(event.ff_release_images, fedfind.release.ModularCompose)

    @mock.patch("fedfind.release.IoTNightly.exists", True)
    @mock.patch("fedfind.release.IoTNightly.all_images", ["foo"])
    def test_candidate_ff_release_iot(self):
        """Straightforward ff_release test case for an IoT 'nightly'
        compose.
        """
        event = wikitcms.event.NightlyEvent(self.site, "33", "RC", "20200513.0", dist="Fedora-IoT")
        assert isinstance(event.ff_release, fedfind.release.IoTNightly)
        assert isinstance(event.ff_release_images, fedfind.release.IoTNightly)


@pytest.mark.usefixtures("fakemwp")
@mock.patch("wikitcms.page.Page.save", autospec=True)
@mock.patch("wikitcms.page.SummaryPage.update_current", autospec=True)
@mock.patch("wikitcms.page.ValidationPage.update_current", autospec=True)
@mock.patch("wikitcms.event.ValidationEvent.update_current", autospec=True)
@mock.patch(
    "test_event.wk.Wiki.get_testtypes",
    return_value=["Installation", "Base", "Server", "Cloud", "Desktop"],
)
@mock.patch("fedfind.release.BranchedNightly.cid", "Fedora-27-20171104.n.0")
@mock.patch(
    "fedfind.helpers.download_json",
    return_value={
        "arguments": {},
        "count": 1,
        "pages": 1,
        "total": 1,
        "raw_messages": [
            {
                "msg": {
                    "architecture": "x86_64",
                    "compose": "Fedora-27-20171104.n.0",
                    "destination": "eu-west-2",
                    "extra": {
                        "id": "ami-085e29c4cd80e326d",
                        "virt_type": "hvm",
                        "vol_type": "gp2",
                    },
                }
            }
        ],
    },
)
class TestEventCreate:
    """Tests related to event creation."""

    site = wk.Wiki("fjajah", do_init=False, force_login=False)

    @mock.patch("fedfind.release.BranchedNightly.exists", True)
    @mock.patch(
        "fedfind.release.BranchedNightly.all_images",
        [
            {
                "arch": "x86_64",
                "format": "iso",
                "path": (
                    "Workstation/x86_64/iso/" "Fedora-Workstation-Live-x86_64-27-2011104.n.0.iso"
                ),
                "subvariant": "Workstation",
                "type": "live",
                "url": (
                    "https://kojipkgs.fedoraproject.org/compose/branched/"
                    "Fedora-27-20171104.n.0/Workstation/x86_64/iso/"
                    "Fedora-Workstation-Live-x86_64-27-2011104.n.0.iso"
                ),
            }
        ],
    )
    def test_event_create(self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave):
        """Test normal event creation."""
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        event.create()
        # we should save 5 test pages, plus the summary page,
        # download page, AMI page and two category pages
        assert fakepagesave.call_count == 10
        # we should call update_current for all 5 test pages
        assert fakepageup.call_count == 5
        # and also for the Summary page
        assert fakesumup.call_count == 1
        # we should call update_current for the event itself
        assert fakeevup.call_count == 1
        # verify we set createonly to True by default
        for call in fakepagesave.call_args_list:
            assert call[1]["createonly"] is True

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_event_create_no_images(
        self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave
    ):
        """Test event creation where no images are available. This
        should succeed, but not create a download page.
        """
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        event.create()
        # we should save 5 test pages, plus the summary page and
        # two category pages and AMI page - but no download page
        assert fakepagesave.call_count == 9
        # we should call update_current for all 5 test pages
        assert fakepageup.call_count == 5
        # we should call update_current for the event itself
        assert fakeevup.call_count == 1

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_event_create_check(
        self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave
    ):
        """Test event creation with check=True."""
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        # first, let's say the pages had no content, so creation
        # should go ahead...
        with mock.patch("mwclient.page.Page.text", return_value=""):
            event.create(check=True)
        # we should save 5 test pages, plus the summary page and
        # two category pages and AMI page - but no download page
        assert fakepagesave.call_count == 9
        # now, let's say the pages *do* have content, we should get
        # an exception now...
        fakepagesave.reset_mock()
        with mock.patch("mwclient.page.Page.text", return_value="foobar"):
            with pytest.raises(ValueError):
                event.create(check=True)
        assert fakepagesave.call_count == 0

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_event_create_testtypes(
        self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave
    ):
        """Test event creation for a specified set of test types."""
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        event.create(testtypes=["Installation", "Server"])
        # we should save 2 test pages, plus the summary page and
        # two category pages and AMI page - but no download page
        assert fakepagesave.call_count == 6
        # we should call update_current for both test pages
        assert fakepageup.call_count == 2
        # we should call update_current for the event itself
        assert fakeevup.call_count == 1
        # check correct handling of garbage list of testtypes
        with pytest.raises(ValueError):
            event.create(testtypes=["foo", "bar"])

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_event_create_force(
        self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave
    ):
        """Test event creation with force=True."""
        # we only need to test that we properly set createonly=None
        # here, we don't need to confirm or mock its effects
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        event.create(force=True)
        for call in fakepagesave.call_args_list:
            assert call[1]["createonly"] is None

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_event_create_existing(
        self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave
    ):
        """Test the handling when a page we try to create during
        event creation already exists. The expected behaviour is that
        we just log the issue and continue.
        """
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        fakepagesave.side_effect = mwclient.errors.APIError("articleexists", "Article exists", {})
        event.create()
        # but if the error is something *else*, we should raise it
        fakepagesave.side_effect = mwclient.errors.APIError("frobbled", "Article is frobbled", {})
        with pytest.raises(mwclient.errors.APIError):
            event.create()
        fakepagesave.reset_mock()

    @mock.patch("fedfind.release.BranchedNightly.exists", False)
    @mock.patch("fedfind.release.BranchedNightly.all_images", [])
    def test_event_create_nocurrent(
        self, fakejson, faketypes, fakeevup, fakepageup, fakesumup, fakepagesave
    ):
        """Test that setting current=False correctly skips current
        redirect updates.
        """
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        event.create(current=False)
        # we should save 5 test pages, plus the summary page and
        # two category pages and AMI page - but no download page
        assert fakepagesave.call_count == 9
        # we should not call update_current for test pages
        assert fakepageup.call_count == 0
        # we should not call update_current for the event itself
        assert fakeevup.call_count == 0


class TestEventOther:
    """Tests for properties of Event instances and other methods."""

    site = wk.Wiki("fjajah", do_init=False, force_login=False)

    @mock.patch("mwclient.listing.GeneratorList.__next__")
    def test_event_result_pages(self, fakenext, fakemwp):
        """Test for result_pages property."""
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        # we need to mock the output of the generator the method uses
        # it's an 'allresults' generator (so it uses the Test Results
        # namespace) and the prefix should be "Fedora 27 Branched
        # 20171104.n.0 ", so what we're expecting to get back is all
        # the actual result pages plus the summary page, and we have
        # to test the method correctly filters out the summary
        evp1 = mwclient.page.Page(self.site, "Test Results:Fedora 27 Branched 20171104.n.0 Base")
        evp2 = mwclient.page.Page(self.site, "Test Results:Fedora 27 Branched 20171104.n.0 Cloud")
        evs = mwclient.page.Page(self.site, "Test Results:Fedora 27 Branched 20171104.n.0 Summary")
        fakenext.side_effect = [evp1, evp2, evs, StopIteration]
        rps = event.result_pages
        assert len(rps) == 2
        (rp1, rp2) = rps
        assert isinstance(rp1, wikitcms.page.NightlyPage)
        assert isinstance(rp2, wikitcms.page.NightlyPage)
        assert rp1.testtype == "Base"
        assert rp2.testtype == "Cloud"
        assert rp1.dist == rp2.dist == "Fedora"
        # modular
        event = wikitcms.event.NightlyEvent(
            self.site, "27", "Branched", "20171123.n.1", dist="Fedora-Modular"
        )
        evp1 = mwclient.page.Page(
            self.site, "Test Results:Fedora-Modular 27 Branched 20171123.n.1 Base"
        )
        evp2 = mwclient.page.Page(
            self.site, "Test Results:Fedora-Modular 27 Branched 20171123.n.1 Server"
        )
        evs = mwclient.page.Page(
            self.site, "Test Results:Fedora-Modular 27 Branched 20171123.n.1 Summary"
        )
        fakenext.side_effect = [evp1, evp2, evs, StopIteration]
        rps = event.result_pages
        assert len(rps) == 2
        (rp1, rp2) = rps
        assert isinstance(rp1, wikitcms.page.NightlyPage)
        assert isinstance(rp2, wikitcms.page.NightlyPage)
        assert rp1.testtype == "Base"
        assert rp2.testtype == "Server"
        assert rp1.dist == rp2.dist == "Fedora-Modular"

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_event_update_current(self, fakesave, fakemwp):
        """Test for update_current() method."""
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        event.update_current()
        assert fakesave.call_args[0][1] == (
            "{{tempdoc}}\n<onlyinclude>{{#switch: {{{1|full}}}\n| full = 27 Branched 20171104.n.0"
            "\n| release = 27\n| milestone = Branched\n| compose =\n| date = 20171104.n.0\n"
            "}}</onlyinclude>\n[[Category: Fedora Templates]]"
        )
        assert fakesave.call_args[0][2] == "relval: update to current event"
        assert fakesave.call_args[1]["createonly"] is None
        assert fakesave.call_args[0][0].name == "Template:CurrentFedoraCompose"
        # modular, non-nightly
        event = wikitcms.event.ComposeEvent(self.site, "27", "Beta", "1.5", dist="Fedora-Modular")
        event.update_current()
        assert fakesave.call_args[0][1] == (
            "{{tempdoc}}\n<onlyinclude>{{#switch: {{{1|full}}}\n| full = 27 Beta 1.5"
            "\n| release = 27\n| milestone = Beta\n| compose = 1.5\n| date =\n"
            "}}</onlyinclude>\n[[Category: Fedora Templates]]"
        )
        assert fakesave.call_args[0][2] == "relval: update to current event"
        assert fakesave.call_args[1]["createonly"] is None
        assert fakesave.call_args[0][0].name == "Template:CurrentFedora-ModularCompose"
        # IoT 'nightly'
        event = wikitcms.event.NightlyEvent(self.site, "33", "RC", "20200513.0", dist="Fedora-IoT")
        event.update_current()
        assert fakesave.call_args[0][1] == (
            "{{tempdoc}}\n<onlyinclude>{{#switch: {{{1|full}}}\n| full = 33 RC 20200513.0"
            "\n| release = 33\n| milestone = RC\n| compose =\n| date = 20200513.0\n"
            "}}</onlyinclude>\n[[Category: Fedora Templates]]"
        )
        assert fakesave.call_args[0][2] == "relval: update to current event"
        assert fakesave.call_args[1]["createonly"] is None
        assert fakesave.call_args[0][0].name == "Template:CurrentFedora-IoTCompose"

    def test_event_from_page(self, fakemwp):
        """Test for from_page classmethod."""
        pg = self.site.pages["Test Results:Fedora 27 Branched 20171104.n.0 Base"]
        ev = wikitcms.event.NightlyEvent.from_page(pg)
        assert ev.version == "27 Branched 20171104.n.0"
        assert ev.dist == "Fedora"
        # modular
        pg = self.site.pages["Test Results:Fedora-Modular 27 Branched 20171123.n.1 Base"]
        ev = wikitcms.event.NightlyEvent.from_page(pg)
        assert ev.version == "27 Branched 20171123.n.1"
        assert ev.dist == "Fedora-Modular"
        # IoT
        pg = self.site.pages["Test Results:Fedora-IoT 33 RC 20200513.0 General"]
        ev = wikitcms.event.NightlyEvent.from_page(pg)
        assert ev.version == "33 RC 20200513.0"
        assert ev.dist == "Fedora-IoT"

    def test_event_category_page(self, fakemwp):
        """Test for category_page property."""
        # compose
        event = wikitcms.event.ComposeEvent(self.site, "27", "Beta", "1.2")
        cat = event.category_page
        assert isinstance(cat, wikitcms.listing.ValidationCategory)
        # this is a sufficient proxy that we got the right thing
        assert cat.checkname == "Category:Fedora 27 Beta Test Results"
        # modular compose
        event = wikitcms.event.ComposeEvent(self.site, "27", "Beta", "1.5", dist="Fedora-Modular")
        cat = event.category_page
        assert cat.checkname == "Category:Fedora-Modular 27 Beta Test Results"
        # nightly
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        cat = event.category_page
        assert cat.checkname == "Category:Fedora 27 Nightly Test Results"
        # modular nightly
        event = wikitcms.event.NightlyEvent(
            self.site, "27", "Branched", "20171123.n.1", dist="Fedora-Modular"
        )
        cat = event.category_page
        assert cat.checkname == "Category:Fedora-Modular 27 Nightly Test Results"
        # IoT
        event = wikitcms.event.NightlyEvent(self.site, "33", "RC", "20200513.0", dist="Fedora-IoT")
        cat = event.category_page
        assert cat.checkname == "Category:Fedora-IoT 33 Nightly Test Results"

    # yes, I looked up the correct date. don't make fun of me!
    @mock.patch("wikitcms.page.ComposePage.creation_date", "20170922")
    def test_event_creation_date(self, fakemwp):
        """Test for creation_date property/attribute."""
        # compose
        evp = self.site.pages["Test Results:Fedora 27 Beta 1.2 Base"]
        with mock.patch("wikitcms.event.ComposeEvent.result_pages", [evp]):
            event = wikitcms.event.ComposeEvent(self.site, "27", "Beta", "1.2")
            assert event.creation_date == "20170922"
        with mock.patch("wikitcms.event.ComposeEvent.result_pages", []):
            # this is when the event doesn't exist
            event = wikitcms.event.ComposeEvent(self.site, "27", "Beta", "1.2")
            assert event.creation_date == ""
        # nightly - it's just an attribute here
        event = wikitcms.event.NightlyEvent(self.site, "27", "Branched", "20171104.n.0")
        assert event.creation_date == "20171104"



# -*- coding: utf-8 -*-

# Copyright (C) 2016 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument, too-many-arguments

"""Tests for result.py."""

import wikitcms.result as rs


class TestResults:
    """Tests for find_results()."""

    def _check_result(
        self,
        res,
        status=None,
        user=None,
        bugs=None,
        comment="",
        bot=False,
        transferred=False,
        comment_bugs=None,
    ):
        if bugs is None:
            bugs = []
        if comment_bugs is None:
            comment_bugs = set()
        assert res.status == status
        assert res.user == user
        assert res.bugs == bugs
        assert res.comment == comment
        assert res.bot == bot
        assert res.transferred == transferred
        assert res.comment_bugs == comment_bugs

    def test_find_results_current_template(self):
        """Various common forms of the current template."""

        restext = """
{{result|none}}
{{result|pass}}
{{result|fail|adamwill}}
{{result|warn|adamwill|345234}}
{{result|pass|kparal|6789032}} <ref>Some comment.</ref>
{{result|fail||3456780}}
{{result|pass|previous RC3 run}}
{{result|pass|coconut|bot=true}}
{{result |fail |  rené|372312 |9345671}}
"""
        results = rs.find_results(restext)
        assert len(results) == 9
        self._check_result(results[0])
        self._check_result(results[1], status="pass")
        self._check_result(results[2], status="fail", user="adamwill")
        self._check_result(results[3], status="warn", user="adamwill", bugs=["345234"])
        self._check_result(
            results[4],
            status="pass",
            user="kparal",
            bugs=["6789032"],
            comment="<ref>Some comment.</ref>",
        )
        self._check_result(results[5], status="fail", bugs=["3456780"])
        self._check_result(results[6], status="pass", user="previous RC3 run", transferred=True)
        self._check_result(results[7], status="pass", user="coconut", bot=True)
        self._check_result(results[8], status="fail", user="rené", bugs=["372312", "9345671"])

    def test_find_results_old_template(self):
        """Various common forms of the old testresult template."""
        restext = """
{{testresult/none}}
{{testresult/pass|jlaska}}
{{testresult/fail|jkeating}} <ref>{{bz|517926}}</ref>
{{testresult/fail|jlaska}}  <ref>{{bz|533420}}- broken dep: gnome-python2-gdl-2.25.3-12.fc12.i686 requires libgdl-1.so.2</ref>
{{testresult/warn|jlaska}}  <ref>NEEDS REVIEW - https://fedorahosted.org/pipermail/autoqa-results/2009-November/001966.html</ref>
"""
        results = rs.find_results(restext)
        assert len(results) == 5
        self._check_result(results[0])
        self._check_result(results[1], status="pass", user="jlaska")
        self._check_result(results[2], status="fail", user="jkeating", bugs=["517926"])
        res3comm = (
            "<ref>- broken dep: gnome-python2-gdl-2.25.3-12.fc12.i686 requires "
            "libgdl-1.so.2</ref>"
        )
        self._check_result(
            results[3], status="fail", user="jlaska", bugs=["533420"], comment=res3comm
        )
        res4comm = (
            "<ref>NEEDS REVIEW - https://fedorahosted.org/pipermail/autoqa-results/"
            "2009-November/001966.html</ref>"
        )
        self._check_result(results[4], status="warn", user="jlaska", comment=res4comm)

    def test_find_results_exclusions(self):
        """Test excluding example, 'transferred' and 'bot' results."""
        restext = """
{{result|pass|adamwill}}
{{result|fail|SampleUser}}
{{result|fail|FASName}}
{{result|pass|previous RC2 run}}
{{result|pass|coconut|bot=true}}
"""
        assert len(rs.find_results(restext)) == 3
        assert len(rs.find_results(restext, transferred=False)) == 2
        assert len(rs.find_results(restext, bot=False)) == 2
        assert len(rs.find_results(restext, transferred=False, bot=False)) == 1

    def test_find_results_statuses(self):
        """Test filtering results by status."""
        restext = """
{{result|none}}
{{result|pass}}
{{result|fail|adamwill}}
{{result|pass|coconut|bot=true}}
{{result|warn|kparal}}
{{result|warn|jlaska}}
"""

        assert len(rs.find_results(restext)) == 6
        assert len(rs.find_results(restext, statuses=["pass", "fail"])) == 3
        assert len(rs.find_results(restext, statuses=["pass", "warn"], bot=False)) == 3

    def test_find_results_page_extracts(self):
        """Various tests culled from real test pages, with messy text."""
        # From Fedora_12_Final_RC1_Install. Includes sample results
        # which should be filtered (but the 'nones' cannot be).
        f12text = """
== Key ==

Please use the following format when posting results to this page


{| class="wikitable sortable" border="1"
! Priority !! Test Case !! i386 !! ppc !! x86_64 !! References
|- 
| Tier1
| A sample test case
| {{testresult/none}} <ref>Indicates untested</ref>
| {{testresult/pass|FASName}} <ref>Indicates a test that has passed </ref>
| {{testresult/fail|FASName}} <ref>Indicates a failed test ... see ''Notes/Bug(s)'' for details </ref>
| <references/> 
|-
| Tier2
| Another sample test case
| {{testresult/warn|FASName}} <ref>Indicates an inprogress or test that needs further review </ref>
| {{testresult/none}}
| style="background:lightgrey;"| {{testresult/none}} <ref> Indicates unsupported configuration </ref>
| <references/> 
|-
|}

== Priority ==

The explanation of test case priority is available at [[QA:Fedora_12_Install_Test_Plan#Test_Priority]].

= '''Test Matrix''' =

{| class="wikitable sortable" border="1"
! Priority !! Test Area !! Test Case !! i386 !! ppc !! x86_64 !! References
|-
| Tier1
| Image Sanity
| [[QA:Testcase_Mediakit_ISO_Size]]
| {{testresult/pass|kparal}} 
| {{testresult/pass|kparal}} 
| {{testresult/pass|kparal}} 
| <references/> 
|-
| Tier1
| Image Sanity
| [[QA:Testcase_Mediakit_ISO_Checksums]]
| {{testresult/none}}
| {{testresult/none}} 
| {{testresult/none}} 
| <references/> 
|- 
| Tier3
"""
        results = rs.find_results(f12text)
        assert len(results) == 9
        for res in results[3:6]:
            self._check_result(res, status="pass", user="kparal")
        for res in results[6:8]:
            self._check_result(res)

        # extracts from Fedora_16_Final_RC1_Install. Result split
        # across lines, multiple results on one line.
        f16text = """
|-
| Alpha
| Install Repository
| [[QA:Testcase_install_repository_Mirrorlist_default]]
| {{result|pass|rmarko}} <ref>{{bz|744463}} Final blocker?
</ref>
| {{result|pass|rmarko}}
|-
| Alpha
| Install Repository
| [[QA:Testcase_install_repository_DVD_default]]
| {{result|pass|rmarko}}{{result|pass|robatino}}
| {{result|pass|rmarko}}{{result|pass|robatino}}
|-
"""
        results = rs.find_results(f16text)
        assert len(results) == 6
        self._check_result(
            results[0],
            status="pass",
            user="rmarko",
            comment="<ref>{{bz|744463}} Final blocker?\n</ref>",
            comment_bugs=set(["744463"]),
        )
        for res in (results[1], results[2], results[4]):
            self._check_result(res, status="pass", user="rmarko")
        for res in (results[3], results[5]):
            self._check_result(res, status="pass", user="robatino")

        # from Test_Results:Fedora_23_Final_RC3_Installation
        # bot text mixed with regular results
        f23text = """
|-
| Alpha
| [[QA:Testcase_partitioning_guided_empty]]
| {{result|pass|coconut|bot=true}}{{result|pass|juliuxpigface}}
| {{result|pass|coconut|bot=true}}
| {{result|none}}
|-
| Alpha
| [[QA:Testcase_partitioning_guided_delete_all]]
| {{result|pass|roshi}}{{result|pass|coconut|bot=true}}{{result|pass|juliuxpigface}}
| {{result|pass|coconut|bot=true}}
| {{result|none}}
|-
| Alpha
| [[QA:Testcase_partitioning_guided_multi_select]]
| {{result|pass|coconut|bot=true}}
| {{result|pass|coconut|bot=true}}
| style="background:lightgrey;"|
|-
"""
        results = rs.find_results(f23text)
        resexc = rs.find_results(f23text, bot=False)
        assert len(results) == 11
        assert len(resexc) == 5
        for res in (results[0], results[2], results[5], results[7], results[9], results[10]):
            self._check_result(res, status="pass", user="coconut", bot=True)
        for res in (results[3], results[8]):
            self._check_result(res)
        for res in (results[1], results[6]):
            self._check_result(res, status="pass", user="juliuxpigface")
        self._check_result(results[4], status="pass", user="roshi")

    def test_find_results_by_row(self):
        """Tests for find_results_by_row."""
        # extracts from Test_Day:2011-02-23_Radeon
        # example results (should be filtered), various username styles
        # one result with a username included added to make coverage
        # happy
        restext = """
=== Main tests ===

{|
! User
! Smolt Profile
! [[QA:Testcase_radeon_basic|Basic test]]
! [[QA:Testcase_radeon_gnome3|GNOME 3 start]]
! [[QA:Testcase_radeon_dpms|DPMS]]
! [[QA:Testcase_radeon_xv|XVideo]]
! [[QA:Testcase_radeon_rotate|Rotation]]
! [[QA:Testcase_radeon_restartx|X restart]]
! [[QA:Testcase_radeon_rendercheck|Rendercheck]]
! [[QA:Testcase_radeon_glx|GLX]]
! [[QA:Testcase_radeon_fastuserswitch|User switch]]
! [[QA:Testcase_radeon_vt|VT switch]]
! [[QA:Testcase_radeon_suspend|Suspend]]
! [[QA:Testcase_radeon_multihead|Multihead]]
! Comments
|-
| [[User:Example | Example user]] 
| [http://www.smolts.org/client/show/pub_84465125-1350-4f83-87b9-5f16f7430eb8 HW]
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| N/A
| {{result|fail||123456}}
| {{result|warn||234567}}
| <references/>
|-
| [[Denis Kurochkin]] 
| [http://www.smolts.org/client/show/pub_76e7df41-e4bb-406b-a9d2-c70564d479b7 HW]
| {{result|warn}} <ref group="long">Resolution detected as 800x600. System - system settings - display - mirror display is setted on by default. While it setted on I can not change resolution</ref>
| {{result|warn}} <ref>UI bugs like this [http://i52.tinypic.com/2qixm4y.png screenshot] </ref>
| {{result|pass}}
| {{result|pass}}
| {{result|inprogress}}
| {{result|inprogress}}
| {{result|inprogress}}
| {{result|inprogress}}
| {{result|inprogress}}
| {{result|inprogress}}
| {{result|inprogress}}
| {{result|inprogress}}
| <references/>
|-
| MadRouter
| [http://www.smolts.org/client/show/pub_7a5f902f-a866-4810-9a14-e5b3a9702cd6 HW]
| {{result|pass}}
| {{result|warn}}<ref>gnome shell does not start after first boot, start successfully after next boots</ref>
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}<ref>[[File:Radeon_renderchecklog_madrouter.gz]]</ref>
| {{result|pass}}
| {{result|pass}}<ref>work fine from gnome 3 shell, N/A from gnome desktop</ref>
| {{result|pass}}
| {{result|warn}}<ref>shut down instead of suspend from gnome 3 shell, but work fine from gnome desktop</ref>
| {{result|pass}}
| <references/>
|-
| [[User:Mira | Mira]] 
| [http://www.smolts.org/client/show/pub_94b448bc-4277-4c30-89cc-0c2b7047f9ca HW]
| {{result|pass}}
| {{result|warn}} <ref>Always after 1st log-in Gnome in fallback mode, after log-out log-in Gnome shell started</ref>
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|pass}}
| {{result|fail}} <ref>System didn't suspend successfully, it just simply switched off</ref>
| {{result|warn|foobar}}
| N/A
| <references/>
|-
"""
        results = rs.find_results_by_row(restext)
        assert len(results) == 36
        for res in results[0:12]:
            assert res.user == "[[denis kurochkin]]"
        for res in results[12:24]:
            assert res.user == "madrouter"
        for res in results[24:34]:
            assert res.user == "mira"
        assert results[35].user == "foobar"

    def _check_resultrow(
        self, row, testcase, milestone, columns, results, section="", secid=0, name=None
    ):
        assert row.testcase == testcase
        assert row.milestone == milestone
        assert row.columns == columns
        assert row.section == section
        assert row.secid == secid
        for key in row.results.keys():
            row.results[key] = len(row.results[key])
        assert row.results == results
        if not name:
            name = testcase
        assert row.name == name

    def test_find_resultrows_simple(self):
        """find_resultrows: simple table."""
        text = """
==== Virtualization ====
</div>
{| class="wikitable sortable collapsible" width=100%
|- 
!style="width:10em"|Milestone !! Test Case !! x86<ref>'x86' means 'x86_64 or i386'</ref> BIOS !! x86 UEFI !! ARM
|-
| Beta
| [[QA:Testcase_Install_to_Previous_KVM]]
| {{result|pass|previous RC6 run}}{{result|pass|pschindl}}
| {{result|warn|previous RC6 run}}<ref>openQA's UEFI stuff has issues</ref>
| {{result|none}}
|-
| Beta
| [[QA:Testcase_Install_to_Current_KVM]]
| {{result|pass|previous RC6 run}}
| {{result|pass|previous RC3 run}}
| {{result|none}}
|-
| Final
| [[QA:Testcase_Boot_Methods_Xen_Para_Virt]]
| {{result|pass|previous RC3 run}}
| {{result|none}}
| {{result|none}}
|-
|}
"""
        rows = rs.find_resultrows(text)
        assert len(rows) == 3
        cols = ["Milestone", "Test Case", "x86 BIOS", "x86 UEFI", "ARM"]
        self._check_resultrow(
            rows[0],
            "QA:Testcase_Install_to_Previous_KVM",
            "Beta",
            cols,
            {"x86 BIOS": 2, "x86 UEFI": 1, "ARM": 1},
        )
        self._check_resultrow(
            rows[1],
            "QA:Testcase_Install_to_Current_KVM",
            "Beta",
            cols,
            {"x86 BIOS": 1, "x86 UEFI": 1, "ARM": 1},
        )
        self._check_resultrow(
            rows[2],
            "QA:Testcase_Boot_Methods_Xen_Para_Virt",
            "Final",
            cols,
            {"x86 BIOS": 1, "x86 UEFI": 1, "ARM": 1},
        )

    def test_find_resultrows_names(self):
        """find_resultrows: test instances by name."""
        text = """
{| class="wikitable sortable collapsible" width=100%
|-
!style="width:10em"|Milestone !! Image !! i386 !! x86_64 !! UEFI
|-
| Alpha
| [[QA:Testcase_Boot_default_install|Workstation live]]
| {{result|pass|coconut|bot=true}}
| {{result|pass|adamwill}}{{result|pass|coconut|bot=true}}
| {{result|pass|adamwill}}{{result|pass|pschindl}}<ref>burned on dvd</ref>{{result|pass|coconut|bot=true}}
|-
| Alpha
| [[QA:Testcase_Boot_default_install|Workstation netinst]]
| {{result|pass|kparal}}
| {{result|pass|pschindl}}
| {{result|pass|pschindl}}
|-
| Alpha
| [[QA:Testcase_Boot_default_install|Server netinst]]
| {{result|pass|coconut|bot=true}}
| {{result|pass|pschindl}}{{result|pass|coconut|bot=true}}
| {{result|pass|pschindl}}{{result|pass|coconut|bot=true}}
|-
"""
        rows = rs.find_resultrows(text)
        assert len(rows) == 3
        cols = ["Milestone", "Image", "i386", "x86_64", "UEFI"]
        self._check_resultrow(
            rows[0],
            "QA:Testcase_Boot_default_install",
            "Alpha",
            cols,
            {"i386": 1, "x86_64": 2, "UEFI": 3},
            name="Workstation live",
        )
        self._check_resultrow(
            rows[1],
            "QA:Testcase_Boot_default_install",
            "Alpha",
            cols,
            {"i386": 1, "x86_64": 1, "UEFI": 1},
            name="Workstation netinst",
        )
        self._check_resultrow(
            rows[2],
            "QA:Testcase_Boot_default_install",
            "Alpha",
            cols,
            {"i386": 1, "x86_64": 2, "UEFI": 2},
            name="Server netinst",
        )

    def test_find_resultrows_multiline(self):
        """find_resultrows: results split across lines."""
        text = """
{| class="wikitable sortable collapsible" border="1" width=100%
! Milestone !! Test Case  !! Local !! EC2 !! Openstack
|-
| Alpha
| [[QA:Testcase_base_startup]]
| {{result|pass|previous RC3 run}}{{result|pass|roshi}}
| {{result|pass|dustymabe}}ami-f43b489e
| {{result|pass|kevin}}
  {{result|pass|dustymabe}}
|-
|}
"""
        rows = rs.find_resultrows(text)
        assert len(rows) == 1
        self._check_resultrow(
            rows[0],
            "QA:Testcase_base_startup",
            "Alpha",
            ["Milestone", "Test Case", "Local", "EC2", "Openstack"],
            {"Local": 2, "EC2": 1, "Openstack": 2},
        )

    def test_find_resultrows_text_between_tables(self):
        """find_resultrows: check text containing test case names between
        tables is not parsed into a garbage row.
        """
        text = """
{| class="wikitable sortable collapsible" width=100%
|-
!style="width:10em"|Milestone !! Test Case !! Server !! Workstation !! Spins !! ARM disks !! Cloud
|-
| Alpha / Final
| [[QA:Testcase_Mediakit_Checksums]]
| {{result|pass|kparal}}{{result|pass|robatino}}
| {{result|pass|kparal}}{{result|pass|robatino}}
| {{result|pass|robatino}}
| {{result|pass|robatino}}
| {{result|pass|robatino}}
|-
|}
<references/>
<div style="background-color: lightgreen; text-align: center; border-style: solid; border-width: thin">

==== Default boot and install ====
</div>
{{admon/important|Bare metal testing required|Running these tests in a virtualized environment is valuable, but the test must pass on bare metal using a physical install medium for each release-blocking image in each test environment at least for the final release candidate image (i.e. the released image) for each milestone.}}

{{admon/note|Single test table|In all of these tests, the test case used is [[QA:Testcase_Boot_default_install]]. That is where the links point. The same test needs to be run for multiple images and target platforms. Note that the non-installer-based ARM disk images are covered by the later [[#ARM disk images]] section.}}
{| class="wikitable sortable collapsible" width=100%
|-
!style="width:10em"|Milestone !! Image !! i386 !! x86_64 !! UEFI
|-
| Alpha
| [[QA:Testcase_Boot_default_install|Workstation live]]
| {{result|pass|coconut|bot=true}}
| {{result|pass|adamwill}}{{result|pass|coconut|bot=true}}
| {{result|pass|adamwill}}{{result|pass|pschindl}}<ref>burned on dvd</ref>{{result|pass|coconut|bot=true}}
|-
|}
"""
        rows = rs.find_resultrows(text)
        assert len(rows) == 2

    def test_find_resultrows_column_titles_no_separator(self):
        """Test find_resultrows works OK when there is no separator
        before the column titles.
        """
        text = """
{| class="wikitable sortable" border="1"
! Priority !! Test Area !! Test Case !! i386 !! ppc !! x86_64 !! References
|- 
| Tier1
| Image Sanity
| [[QA:Testcase_Mediakit_ISO_Size]]
| {{testresult/none}}
| {{testresult/none}}
| {{testresult/none}}
| <references/> 
|-
"""
        rows = rs.find_resultrows(text)
        assert rows[0].columns == [
            "Priority",
            "Test Area",
            "Test Case",
            "i386",
            "ppc",
            "x86_64",
            "References",
        ]

    def test_find_resultrows_column_titles_sanitize(self):
        """Test the cleanups find_resultrows does on column names."""
        text = """
{| class="wikitable sortable" border="1"
|-
! Priority !! Test Area !! Test Case !! ''i386'' !! [[Somewhere|ppc]] !! x86_64 <ref>some comment</ref> !! References
|- 
| Tier1
| Image Sanity
| [[QA:Testcase_Mediakit_ISO_Size]]
| {{testresult/none}}
| {{testresult/none}}
| {{testresult/none}}
| <references/> 
|-
"""
        rows = rs.find_resultrows(text)
        assert rows[0].columns == [
            "Priority",
            "Test Area",
            "Test Case",
            "i386",
            "ppc",
            "x86_64",
            "References",
        ]

    def test_result_string_template(self):
        """Test string representation and result_template of Result
        instances."""
        res = rs.Result()
        assert str(res) == "Result placeholder - {{result|none}}"
        assert res.result_template == "{{result|none}}"

        res = rs.Result(status="pass")
        assert str(res) == "Result: Pass"
        assert res.result_template == "{{result|pass}}"

        res = rs.Result(status="fail", user="adamwill")
        assert str(res) == "Result: Fail from adamwill"
        assert res.result_template == "{{result|fail|adamwill}}"

        res = rs.Result(status="warn", user="rené", bugs=["435678", "1340987"])
        assert str(res) == "Result: Warn from rené, bugs: 435678, 1340987"
        assert res.result_template == "{{result|warn|rené|435678|1340987}}"

        res = rs.Result(
            status="pass", user="kparal", comment="some comment <ref>some reference</ref>"
        )
        assert str(res) == "Result: Pass from kparal, comment: some comment some reference"
        assert res.result_template == "{{result|pass|kparal}}some comment <ref>some reference</ref>"

        res = rs.Result.from_result_template("{{result|pass|previous RC3 run}}")
        assert str(res) == "Result: Pass transferred: previous RC3 run"
        assert res.result_template == "{{result|pass|previous RC3 run}}"

        res = rs.Result(status="pass", user="coconut", bot=True)
        assert str(res) == "BOT Result: Pass from coconut"
        assert res.result_template == "{{result|pass|coconut|bot=true}}"

    def test_from_result_template_bug_comment(self):
        """Test from_result_template handling of bug numbers with
        comment refs.
        """
        res = rs.Result.from_result_template("{{result|fail|bob|1034592#c8}}")
        assert res.status == "fail"
        assert res.user == "bob"
        assert res.bugs == ["1034592"]

    def test_resultrow_matches(self):
        """Test the 'matches' method of ResultRows."""
        text = """
{| class="wikitable sortable" border="1"
|-
! Priority !! Test Area !! Test Case !! ''i386'' !! [[Somewhere|ppc]] !! x86_64 <ref>some comment</ref> !! References
|- 
| Tier1
| Image Sanity
| [[QA:Testcase_Mediakit_ISO_Size]]
| {{testresult/none}}
| {{testresult/none}}
| {{testresult/none}}
| <references/> 
|-
"""
        row1 = rs.find_resultrows(text)[0]
        row2 = rs.find_resultrows(text)[0]
        row1.secid = row2.secid = 1
        assert row1.matches(row2)
        row2.name = "foo"
        assert not row1.matches(row2)
        row2.name = row1.name
        row2.secid = 2
        assert not row1.matches(row2)
        row2.secid = row1.secid
        row2.origtext = "foo"
        assert not row1.matches(row2)
        row2.origtext = row1.origtext
        row2.testcase = "foo"
        assert not row1.matches(row2)
        # test with a different type of object
        notarow = "foo"
        assert not row1.matches(notarow)


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2020 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument
# pylint: disable=invalid-name

"""Tests for release.py."""

from unittest import mock

import wikitcms.release
import wikitcms.wiki


class TestRelease:
    """Tests for Release instances."""

    # the "fjajah" is just to make sure we're running offline; if I
    # screw up and add a test that hits the network it"ll cause the
    # tests to hang/fail instead of succeeding a bit slower
    site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)

    def test_release_init(self):
        """Tests for Release.__init__."""
        rl = wikitcms.release.Release("32", self.site)
        # __init__ stuff
        assert rl.release == "32"
        assert rl.dist == "Fedora"
        assert rl.category_name == "Category:Fedora 32 Test Results"
        # modular
        rl = wikitcms.release.Release("27", self.site, dist="Fedora-Modular")
        assert rl.release == "27"
        assert rl.dist == "Fedora-Modular"
        assert rl.category_name == "Category:Fedora-Modular 27 Test Results"

    def test_testday_pages(self, fakemwp):
        """Test for Release.testday_pages. There is actually quite a
        lot going on in this small method, but most of the hard work
        is tested in mwclient upstream and in the listing tests; here
        we can just assume the generator magic works right.
        """
        rl = wikitcms.release.Release("32", self.site)
        tdp = self.site.pages["Test Day:2016-10-24 Cloud"]
        tdp2 = self.site.pages["Test Day:2013-11-05 Printing"]
        other = self.site.pages["Somepage"]
        with mock.patch.object(
            self.site, "pages", {"Category:Fedora 32 Test Days": [tdp, tdp2, other]}
        ):
            assert rl.testday_pages == [tdp, tdp2]

    def test_milestone_pages(self, fakemwp):
        """Test for Release.milestone_pages. Similar to testday_pages,
        we just trust the generator magic here and test it elsewhere.
        """
        rl = wikitcms.release.Release("32", self.site)
        vp1 = self.site.pages["Test Results:Fedora 32 Branched 20200212.n.1 Desktop"]
        vp2 = self.site.pages["Test Results:Fedora 32 Beta 1.2 Base"]
        other = self.site.pages["Somepage"]
        with mock.patch("wikitcms.wiki.Wiki.walk_category", return_value=[vp1, vp2, other]):
            assert list(rl.milestone_pages()) == [vp1, vp2]



# Copyright (C) 2015 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Defines custom exceptions used by wikitcms."""


class NoPageError(Exception):
    """Page does not exist."""


class NotFoundError(Exception):
    """Requested thing wasn't found."""


class TooManyError(Exception):
    """Found too many of the thing you asked for."""


# this inherits from ValueError as the things that raise this may
# previously have passed along a ValueError from fedfind
class FedfindNotFoundError(ValueError, NotFoundError):
    """Couldn't find a fedfind release (probably the fedfind release
    that matches an event).
    """


# vim: set textwidth=100 ts=8 et sw=4:
